(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DpTrans
imports "Q-Restricted_Rewriting"
begin

text {*
A \emph{dependency pair problem} (DP problem, for short) is a tuple consisting
of three components. In the following, we give a high-level (and probably
not entirely correct) description of each component:
\begin{enumerate}
\item
First, we have the \emph{(dependency) pairs} (usually denoted by @{term P} or
@{term \<P>}), accounting for (mutually) recursive dependencies between functions.
\item
Next, there is the set of \emph{(restricting) patterns} (usually denoted by @{term Q}
or @{term \<Q>}). The arguments of each @{term \<R>}-redex have to be in NF with
respect to @{term \<Q>}, before a contraction is possible.
\item
Finally, we have the \emph{rules} (usually denoted by @{term R} or @{term \<R>},
responsible for evaluating function arguments.
\end{enumerate}
*}
type_synonym ('f, 'v) dpp =
  "('f, 'v) trs \<times> ('f, 'v) term set \<times> ('f, 'v) trs"

fun
  ichain ::
    "('f, 'v) dpp \<Rightarrow> (nat \<Rightarrow> ('f, 'v) term) \<Rightarrow> (nat \<Rightarrow> ('f, 'v) term) \<Rightarrow>
     (nat \<Rightarrow> ('f, 'v) subst) \<Rightarrow> bool"
where
  "ichain (P, Q, R) s t \<sigma> = (
    (\<forall>i. (s i,t i) \<in> P) \<and> 
    (\<forall>i. (t i \<cdot> \<sigma> i, s (Suc i) \<cdot> \<sigma> (Suc i)) \<in> (qrstep Q R)^*) \<and>
    (\<forall>i. s i \<cdot> \<sigma> i \<in> NF_terms Q))"

text {*
A \emph{minimal infinite chain} is an infinite chain where additionally all @{text \<sigma>}-instances
of terms in the sequence~@{text t} are terminating w.r.t.~the system~@{text R}.
*}
fun
  min_ichain ::
    "('f, 'v) dpp \<Rightarrow> (nat \<Rightarrow> ('f, 'v) term) \<Rightarrow> (nat \<Rightarrow> ('f, 'v) term)
    \<Rightarrow> (nat \<Rightarrow> ('f, 'v) subst) \<Rightarrow> bool"
where
  "min_ichain (P, Q, R) s t \<sigma> = (
    ichain (P, Q, R) s t \<sigma> \<and> (\<forall>i. SN_elt (qrstep Q R) (t i \<cdot> \<sigma> i)))"

fun
  initial_dpp :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f, 'v) term set \<Rightarrow> ('f, 'v) trs  \<Rightarrow> ('f, 'v) dpp"
where
  "initial_dpp shp Q R = (DP shp (useful_rules Q R), Q, R)"

text {*
The set of minimally nonterminating terms with respect to a relation @{term "R"}.
*}
definition
  Tinf :: "('f, 'v) trs \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "Tinf R t \<equiv> \<not> SN_elt R t \<and> (\<forall>s\<lhd>t. SN_elt R s)"

lemma not_SN_imp_subt_Tinf:
  assumes "\<not> SN_elt R s" shows "\<exists>t\<unlhd>s. Tinf R t"
proof -
  let ?S = "{t | t. s \<unrhd> t \<and> \<not> SN_elt R t}"
  from assms have s: "s \<in> ?S" by auto
  from mp[OF spec[OF spec[OF SN_imp_minimal[OF SN_supt]]] s]
  obtain t where st: "s \<unrhd> t" and nSN: "\<not> SN_elt R t"
    and min: "\<forall> u. (t,u) \<in> supt \<longrightarrow> u \<notin> ?S" by auto
  have "Tinf R t" unfolding Tinf_def 
  proof (simp add: nSN, intro allI impI)
    fix u
    assume u: "t \<rhd> u" 
    from u st have "s \<rhd> u" using supteq_supt_trans by auto
    with min u show "SN_elt R u" by auto
  qed
  with st show ?thesis by auto
qed

lemma Tinf_imp_SN_elt_nrqrstep:
  assumes Tinf: "Tinf (qrstep Q R) t"
  shows "SN_elt (nrqrstep Q R) t"
proof (cases t)
  case (Var x)
  {
    fix s
    assume "(t, s) \<in> nrqrstep Q R"
    then obtain l C \<sigma> where "C \<noteq> \<box>" and "t = C\<langle>l \<cdot> \<sigma>\<rangle>" by auto
    with Var have False by (cases C) auto
  }
  thus ?thesis unfolding SN_elt_def by auto
next
  case (Fun f ss)
  with Tinf have SN: "\<forall>s\<in>set ss. SN_elt (qrstep Q R) s"
    unfolding Tinf_def using supt.arg[of t ss] by auto
  show ?thesis
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain S where start: "S 0 = Fun f ss"
      and steps: "\<forall>i. (S i, S (Suc i)) \<in> nrqrstep Q R"
      unfolding SN_elt_def using Fun by force
    let ?P = "\<lambda>i l r g bef C aft \<sigma>. (l, r) \<in> R \<and> S i = Fun g (bef @ C\<langle>l\<cdot>\<sigma>\<rangle> # aft) \<and>
      S (Suc i) = Fun g (bef @ C\<langle>r\<cdot>\<sigma>\<rangle> # aft) \<and> (\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q)"
    have "\<forall>i. \<exists> l r g bef C aft \<sigma>. ?P i l r g bef C aft \<sigma>"
    proof
      fix i
      from nrqrstepE[OF steps[THEN spec[of _ i]]]
      obtain l r C \<sigma> where "\<forall>u\<lhd>l \<cdot> \<sigma>. u \<in> NF_terms Q" and "(l, r) \<in> R" and "C \<noteq> \<box>"
        and "S i = C\<langle>l \<cdot> \<sigma>\<rangle>" and "S (Suc i) = C\<langle>r \<cdot> \<sigma>\<rangle>" .
      thus "\<exists>l r g bef C aft \<sigma>. ?P i l r g bef C aft \<sigma>" by (cases C, simp, simp, blast)
    qed
    with choice[OF this] obtain l
      where "\<forall>i. \<exists>r g bef C aft \<sigma>. ?P i (l i) r g bef C aft \<sigma>" by auto
    with choice[OF this] obtain r
      where "\<forall>i. \<exists>g bef C aft \<sigma>. ?P i (l i) (r i) g bef C aft \<sigma>" by auto
    with choice[OF this] obtain g
      where "\<forall>i. \<exists>bef C aft \<sigma>. ?P i (l i) (r i) (g i) bef C aft \<sigma>" by auto
    with choice[OF this] obtain bef
      where "\<forall>i. \<exists>C aft \<sigma>. ?P i (l i) (r i) (g i) (bef i) C aft \<sigma>" by auto
    with choice[OF this] obtain C
      where "\<forall>i. \<exists>aft \<sigma>. ?P i (l i) (r i) (g i) (bef i) (C i) aft \<sigma>" by auto
    with choice[OF this] obtain aft
      where "\<forall>i. \<exists>\<sigma>. ?P i (l i) (r i) (g i) (bef i) (C i) (aft i) \<sigma>" by auto
    with choice[OF this] obtain \<sigma>
      where P: "\<forall>i. ?P i (l i) (r i) (g i) (bef i) (C i) (aft i) (\<sigma> i)" by auto
    let ?gtl = "\<lambda>i. Fun (g i) (bef i @ (C i)\<langle>l i\<cdot>\<sigma> i\<rangle> # aft i)"
    let ?gtr = "\<lambda>i. Fun (g i) (bef i @ (C i)\<langle>r i\<cdot>\<sigma> i\<rangle> # aft i)"
    let ?tl = "\<lambda>i. Fun f (bef i @ (C i)\<langle>l i\<cdot>\<sigma> i\<rangle> # aft i)"
    let ?tr = "\<lambda>i. Fun f (bef i @ (C i)\<langle>r i\<cdot>\<sigma> i\<rangle> # aft i)"
    {
      fix i
      have "?P i (l i) (r i) f (bef i) (C i) (aft i) (\<sigma> i) \<and>
        Suc (length (bef i) + length (aft i)) = length ss"
      proof (induct i)
        case 0
        from P[THEN spec[of _ 0]] have "?gtl 0 = Fun f ss" using start[symmetric] by auto
        with P[THEN spec[of _ 0]] show ?case by auto
      next
        case (Suc i)
        from P[THEN spec[of _ i]] have "?gtr i = S (Suc i)" by auto
        also have "\<dots> = ?gtl (Suc i)" using P[THEN spec[of _ "Suc i"]] by auto
        finally have "length (args (?gtr i)) = length (args (?gtl (Suc i)))" by auto
        hence "length (bef (Suc i)) + length (aft (Suc i)) = length (bef i) + length (aft i)"
          by simp          
        with P[THEN spec[of _ "Suc i"]] Suc show ?case by auto
      qed
    }
    note p = this
    { 
      fix j        
      assume j: "j < length ss"
      {
        fix i
        have "?tl (Suc i) = ?tr i" using p[of i] p[of "Suc i"] by auto
        hence id: "args (?tl (Suc i)) ! j = args (?tr i) ! j" by auto
        have "(args (?tl i) ! j, args (?tl (Suc i)) ! j) \<in> (qrstep Q R)^="
        proof (cases "j < length (bef i)")
          case True
          hence "args (?tl i) ! j = (bef i) ! j" by (simp add: nth_append)
          moreover from True
            have "args (?tl (Suc i)) ! j = (bef i) ! j" by (simp only: id, simp add: nth_append)
          ultimately show ?thesis by auto
        next
          case False
          hence jb: "j \<ge> length (bef i)" by auto
          show ?thesis 
          proof (cases "j = length (bef i)")
            case True
            hence "args (?tl i) ! j = (C i)\<langle>l i\<cdot> \<sigma> i\<rangle>" by (simp add: nth_append)              
            moreover have "args (?tl (Suc i)) ! j = (C i)\<langle>r i\<cdot> \<sigma> i\<rangle>"
              by (simp only: id, simp add: nth_append True)
            ultimately show ?thesis using p[of i] by auto
          next
            case False
            obtain k where "k = j - (Suc (length (bef i)))" by auto
            with False j jb p[of i]
              have j: "j = Suc (length (bef i)) + k" and k: "k < length (aft i)" by auto
            hence l: "args (?tl i) ! j = (aft i) ! k" by (simp add: nth_append)              
            moreover have "args (?tl (Suc i)) ! j = (aft i) ! k"
              by (simp only: id, simp add: nth_append j k)
            ultimately show ?thesis by auto
          qed
        qed
      }
      note arg_steps = this
      let ?ti = "\<lambda>i. args (?tl i) ! j" 
      from arg_steps have arg_steps: "\<forall>i. (?ti i, ?ti (Suc i)) \<in> Id \<union> (qrstep Q R)" by auto
      have compat: "Id O qrstep Q R \<subseteq> qrstep Q R" by auto
      from start p[of 0] have "?ti 0 = ss ! j" by auto
      with j have arg: "?ti 0 \<in> set ss" by auto
      hence SN_start: "SN_elt (qrstep Q R) (?ti 0)" using SN and arg by auto
      have "\<exists>m. \<forall>i \<ge> m. (?ti i, ?ti (Suc i)) \<in> Id - qrstep Q R"
        by (rule non_strict_ending[OF arg_steps compat], rule SN_start)
    }
    hence ending: "\<forall>j. \<exists>m. j < length ss \<longrightarrow>
      (\<forall>i\<ge>m. (args (?tl i) ! j, args (?tl (Suc i)) ! j) \<notin> qrstep Q R)" by blast
    from choice[OF ending] obtain m
      where ending: "\<forall>j < length ss.
      (\<forall>i \<ge> m j. (args (?tl i) ! j, args (?tl (Suc i)) ! j) \<notin> qrstep Q R)" by blast
    obtain ma where ma: "ma = max_f m (length ss)" by auto
    hence ma: "\<forall>i < length ss. m i \<le> ma" using max_f by auto
    with ending have ending: "\<forall>j<length ss. (\<forall>i\<ge>ma.
      (args (?tl i) ! j, args (?tl (Suc i)) ! j) \<notin> qrstep Q R)" by auto
    let ?l = "length (bef ma)"
    from p[of ma] have len: "?l < length ss" by auto
    from p[of ma] have l: "args (?tl ma) ! ?l = (C ma)\<langle>l ma \<cdot> \<sigma> ma\<rangle>" by (simp add: nth_append)
    from p[of ma] p[of "Suc ma"] have "?tl (Suc ma) = ?tr ma" by auto
    hence "args (?tl (Suc ma)) ! ?l = args (?tr ma) ! ?l" by auto
    also have "\<dots> = (C ma)\<langle>r ma \<cdot> \<sigma> ma\<rangle>" using p[of ma] by (simp add: nth_append)
    finally have "(args (?tl ma) ! ?l, args (?tl (Suc ma)) ! ?l) \<in> qrstep Q R"
      using p[of ma] by (simp only: l) auto
    with len ending show False by blast
  qed
qed


lemma Tinf_lemma_1:
  assumes "wwf_qtrs Q R" and "Tinf (qrstep Q R) (Fun f ts)" (is "Tinf (qrstep Q R) ?t")
  shows "\<exists>l r u \<sigma>.
  ((l,r) \<in> useful_rules Q R) \<and> 
  (r \<unrhd> u) \<and> 
  (\<forall>x. u \<noteq> (Var x)) \<and> 
  ((?t,l\<cdot>\<sigma>) \<in> (nrqrstep Q R)^*) \<and> 
  ((l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> rqrstep Q R) \<and> 
  ((r\<cdot>\<sigma>) \<unrhd> (u\<cdot>\<sigma>)) \<and>
  (Tinf (qrstep Q R) (l\<cdot>\<sigma>)) \<and>
  (Tinf (qrstep Q R) (u\<cdot>\<sigma>)) \<and>
  \<not>(l \<rhd> u)"
proof -
  let ?U = "useful_rules Q R"
  let ?qr = "qrstep Q R"
  let ?rqr = "rqrstep Q R"
  let ?nrqr = "nrqrstep Q R"
  from assms have "\<And>s. ?t \<rhd> s \<Longrightarrow> SN_elt ?qr s"
    and "\<not> SN_elt ?qr ?t" unfolding Tinf_def by auto
  from `\<not> SN_elt ?qr ?t` have "\<exists>A. A 0 = ?t \<and> (\<forall>i. (A i,A(Suc i)) \<in> ?qr)"
    unfolding SN_elt_def by auto
  then obtain "A" where "A 0 = ?t" and "\<forall>i. (A i,A(Suc i)) \<in> ?qr" by auto
  hence "!!i. (A i,A(Suc i)) \<in> (?rqr \<union> ?nrqr)" by (auto simp: qrstep_iff_rqrstep_or_nrqrstep)
  hence "(A 0,A 1) \<in> ?rqr \<or> (A 0,A 1) \<in> ?nrqr" by force
  hence "\<exists>i. (A 0,A i) \<in> ?nrqr^* \<and> (A i,A(Suc i)) \<in> ?rqr"
  proof
    assume A: "(A 0,A 1) \<in> ?rqr"
    have "(A 0,A 0) \<in> ?nrqr^*" by auto
    with A show ?thesis by auto
  next
    assume B: "(A 0,A 1) \<in> ?nrqr"
    from `Tinf ?qr ?t` have "SN_elt ?nrqr ?t" by (rule Tinf_imp_SN_elt_nrqrstep)
    from `A 0 = ?t`[symmetric] `SN_elt ?nrqr ?t` have "SN_elt ?nrqr (A 0)" by simp
    from `!!i. (A i,A(Suc i)) \<in> (?rqr \<union> ?nrqr)` `SN_elt ?nrqr (A 0)`
    have "\<exists>i. (A i,A(Suc i)) \<in> ?rqr \<and> (\<forall>j<i. (A j,A(Suc j)) \<in> ?nrqr
      \<and> \<not> (A j,A(Suc j)) \<in> ?rqr)"
      using union_iseq_SN_elt_imp_first_step[where R = ?rqr and S = ?nrqr and t = A] by force
    then obtain i where first_rstep: "(A i,A(Suc i)) \<in> ?rqr"
      and "!!j. j<i \<Longrightarrow> (A j,A(Suc j)) \<in> ?nrqr \<and> \<not> (A j,A(Suc j)) \<in> ?rqr" by auto
    hence "\<forall>j<i. (A j,A(Suc j)) \<in> ?nrqr" by auto
    hence "(A 0,A i) \<in> ?nrqr^*" by (induct i) auto
    with first_rstep show ?thesis by auto
  qed
  then obtain i where "(A 0,A i) \<in> ?nrqr^*" and step: "(A i,A(Suc i)) \<in> ?rqr" by auto
  then obtain l r \<sigma>
    where NF: "\<forall>v\<lhd>l \<cdot> \<sigma>. v \<in> NF_terms Q"
    and "(l, r) \<in> R" and "l \<cdot> \<sigma> = A i" and "r \<cdot> \<sigma> = A (Suc i)" by auto
  with step have "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> ?rqr" by auto
  from `(l, r) \<in> R` only_useful_rules[OF NF] have "(l, r) \<in> ?U"
    unfolding useful_rules_def by auto
  from `A 0 = ?t` and `(A 0,A i) \<in> ?nrqr^*` have "(?t,A i) \<in> ?nrqr^*" by simp
  hence "\<exists>ls. A i = Fun f ls" by (rule nrqrsteps_preserve_root)
  then obtain ls' where "A i = Fun f ls'" by auto
  have A: "\<And>i. (ts ! i, ls' ! i) \<in> ?qr^*" (is "\<And>i. ?P i")
  proof -
    fix j
    from `(?t, A i) \<in> ?nrqr^*` have "(args ?t!j, args (Fun f ls')!j) \<in> ?qr^*"
      unfolding `A i = Fun f ls'` by (rule nrqrsteps_imp_arg_qrsteps)
    thus "?P j" by simp
  qed
  from `(?t,A i) \<in> ?nrqr^*` and `A i = Fun f ls'` have "(?t,Fun f ls') \<in> ?nrqr^*" by simp
  from `l\<cdot>\<sigma> = A i` and `A i = Fun f ls'` have "l\<cdot>\<sigma> = Fun f ls'" by simp
  with `(?t,Fun f ls') \<in> ?nrqr^*` have "(?t,l\<cdot>\<sigma>) \<in> ?nrqr^*" by simp
  from `(?t,Fun f ls') \<in> ?nrqr^*` have numargs: "num_args ?t = num_args (Fun f ls')"
    by (rule nrqrsteps_num_args)
  from `wwf_qtrs Q R` and `(l,r) \<in> R` obtain g ls where "l = Fun g ls"
    using wwf_qtrs_imp_left_fun by best
  hence "l\<cdot>\<sigma> = Fun g (map (\<lambda>t. t\<cdot>\<sigma>) ls)" by simp
  with `l\<cdot>\<sigma> = Fun f ls'` have "g = f" and "ls' = map (\<lambda>t. t\<cdot>\<sigma>) ls" by auto
  hence B: "\<forall>i<length ls'. ls' ! i = (ls ! i) \<cdot> \<sigma>" by auto
  have "\<forall>i < num_args ?t. (ts ! i, (ls ! i) \<cdot> \<sigma>) \<in> ?qr^*"
  proof (intro allI impI)
    fix i assume "i < num_args (Fun f ts)"
    with numargs have "i < length ls'" by simp
    with A[of i] show "(ts ! i, ls ! i \<cdot> \<sigma>) \<in> ?qr^*" using B by simp
  qed
  from `l \<cdot> \<sigma> = Fun f ls'` numargs have "num_args ?t = num_args (l\<cdot>\<sigma>)"  by simp
  with `l\<cdot>\<sigma> = Fun f ls'` and `\<forall>i<length ls'. ls'!i = (ls!i)\<cdot>\<sigma>`
  have "!!i. i < num_args ?t \<Longrightarrow> (ls!i)\<cdot>\<sigma> = ls'!i" by auto
  have "!!i. i < num_args ?t \<Longrightarrow> ?t \<rhd> (ts!i)"
  proof -
    fix i assume "i < num_args ?t"
    hence "ts!i \<in> set ts" by simp
    thus "?t \<rhd> (ts!i)" by (rule supt.arg)
  qed
  with `!!s. ?t \<rhd> s \<Longrightarrow> SN_elt ?qr s` have "!!i. i < num_args ?t \<Longrightarrow> SN_elt ?qr (ts!i)" by auto
  with `\<forall>i<num_args ?t. (ts!i,(ls!i)\<cdot>\<sigma>) \<in> ?qr^*` 
  have "!!i. i < num_args ?t \<Longrightarrow> SN_elt ?qr ((ls!i)\<cdot>\<sigma>)" using steps_preserve_SN_elt[of _ _ ?qr] by blast
  with `!!i. i < num_args ?t \<Longrightarrow> (ls!i)\<cdot>\<sigma> = ls'!i` and `num_args ?t = num_args (l\<cdot>\<sigma>)`
    and `l\<cdot>\<sigma> = Fun f ls'` and `!!i. i < num_args ?t \<Longrightarrow> SN_elt ?qr ((ls!i)\<cdot>\<sigma>)`
  have argsSN: "!!i. i < num_args (Fun f ls') \<Longrightarrow> SN_elt ?qr (ls'!i)" by auto
  have "\<forall>s\<lhd>Fun f ls'. SN_elt ?qr s" 
  proof (intro allI impI)
    fix s
    assume tmp: "Fun f ls' \<rhd> s"
    then obtain C where "C \<noteq> \<box>" and "Fun f ls' = C\<langle>s\<rangle>" by auto
    then obtain bef aft C where id: "Fun f ls' = Fun f (bef @ C\<langle>s\<rangle> # aft)" by (cases C, auto)
    hence len: "length bef < num_args (Fun f ls')" by auto
    with argsSN[OF len] id have "SN_elt ?qr (C \<langle>s\<rangle>)" by auto
    with SN_rel_imp_SN_subt[OF ctxt_closed_qrstep _ ctxt_imp_supteq[of C s]]
    show "SN_elt ?qr s" by auto
  qed
  hence "\<forall>s\<lhd>l\<cdot>\<sigma>. SN_elt ?qr s" unfolding `l\<cdot>\<sigma> = Fun f ls'` by simp 
  let ?B = "\<lambda>j. A (i + j)"
  have "?B 0 = l\<cdot>\<sigma>" unfolding `l\<cdot>\<sigma> = A i` by simp
  have "\<forall>i. (?B i,?B(Suc i)) \<in> ?qr" using `\<forall>i. (A i,A(Suc i)) \<in> ?qr` by simp
  with `?B 0 = l\<cdot>\<sigma>` have "\<not> SN_elt ?qr (l\<cdot>\<sigma>)" unfolding SN_elt_def by best
  with `\<forall>s\<lhd>l\<cdot>\<sigma>. SN_elt ?qr s` have tinf_ls: "Tinf ?qr (l\<cdot>\<sigma>)" unfolding Tinf_def by simp
  have "\<And> x. x \<in> vars_term r \<Longrightarrow> l \<rhd> Var x"
  proof -
    fix x
    assume "x \<in> vars_term r" hence "x \<in> vars_term l" using `(l,r) \<in> ?U` and `wwf_qtrs Q R` unfolding wwf_qtrs_def wwf_rule_def useful_rules_def by auto
    thus "l \<rhd> Var x" using `l = Fun g ls` supteq_Var[OF `x \<in> vars_term l`]
      unfolding supt_supteq_conv by auto
  qed
  hence "\<And> x. x \<in> vars_term r \<Longrightarrow> l \<cdot> \<sigma> \<rhd> Var x \<cdot> \<sigma>"
    using supt_stable[of l _ \<sigma>] unfolding supt_supteq_conv by best
  hence xSN: "\<And> x. x \<in> vars_term r \<Longrightarrow> SN_elt ?qr (Var x \<cdot> \<sigma>)" using tinf_ls unfolding Tinf_def by auto
  let ?B = "\<lambda>j. A (Suc i + j)"
  from `r\<cdot>\<sigma> = A (Suc i)` have "?B 0 = r\<cdot>\<sigma>" by simp
  from `\<forall>i. (A i,A(Suc i)) \<in> ?qr` have "\<forall>i. (?B i,?B(Suc i)) \<in> ?qr" by auto
  with `?B 0 = r\<cdot>\<sigma>` have "\<not> SN_elt ?qr (r\<cdot>\<sigma>)" unfolding SN_elt_def by best
  then obtain t' where "r\<cdot>\<sigma> \<unrhd> t'" and "Tinf ?qr t'" using not_SN_imp_subt_Tinf by auto
  have "\<forall>x \<in> vars_term r. \<not>((Var x)\<cdot>\<sigma> \<unrhd> t')"
  proof
    fix x assume "x \<in> vars_term r"
    hence "SN_elt ?qr (Var x\<cdot>\<sigma>)" using xSN by simp
    show "\<not> Var x\<cdot>\<sigma> \<unrhd> t'"
    proof
      assume "Var x\<cdot>\<sigma> \<unrhd> t'"
      with `SN_elt ?qr (Var x\<cdot>\<sigma>)` have "SN_elt ?qr t'" using SN_rel_imp_SN_subt[OF ctxt_closed_qrstep] by auto
      with `Tinf ?qr t'` show "False" unfolding Tinf_def by auto
    qed
  qed
  with `r\<cdot>\<sigma> \<unrhd> t'` have "\<exists>u. r \<unrhd> u \<and> t' = u\<cdot>\<sigma>" using subt_instance_and_not_subst_imp_subt[where s = r and \<sigma> = \<sigma>] by auto
  then obtain u where "r \<unrhd> u" and "t' = u\<cdot>\<sigma>" by best
  have "\<forall>x. u \<noteq> Var x"
  proof (rule ccontr)
    assume "\<not> (\<forall>x. u \<noteq> Var x)"
    hence "\<exists>x. u = Var x" by simp
    then obtain x where "u = Var x" by best
    with `r \<unrhd> u` have "r \<unrhd> Var x" by auto
    hence "x \<in> vars_term r" using var_subt_imp_subt_in_vars by best
    with xSN `u = Var x` have "SN_elt ?qr (u\<cdot>\<sigma>)" by simp
    with `t' = u\<cdot>\<sigma>` have "SN_elt ?qr t'" by simp
    with `Tinf ?qr t'` show "False" unfolding Tinf_def by simp
  qed
  from `t' = u\<cdot>\<sigma>` and `Tinf ?qr t'` have "Tinf ?qr (u\<cdot>\<sigma>)" by simp
  have "\<not> l \<rhd> u"
  proof
    assume "l \<rhd> u"
    hence "l \<cdot> \<sigma> \<rhd> u \<cdot> \<sigma>" by (rule supt_stable)
    with `Tinf ?qr (l\<cdot>\<sigma>)` have "SN_elt ?qr (u\<cdot>\<sigma>)" unfolding Tinf_def by best
    with `Tinf ?qr (u\<cdot>\<sigma>)` show "False" unfolding Tinf_def by auto
  qed
  from `r \<unrhd> u` have "r\<cdot>\<sigma> \<unrhd> (u\<cdot>\<sigma>)" by (rule supteq_stable)
  from `(l,r) \<in> ?U` and `r \<unrhd> u` and `\<forall>x. u \<noteq> Var x` and `(?t,l\<cdot>\<sigma>) \<in> ?nrqr^*` 
    `(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> ?rqr` and `Tinf ?qr (u\<cdot>\<sigma>)` and `r\<cdot>\<sigma> \<unrhd> (u\<cdot>\<sigma>)` and `Tinf ?qr (l\<cdot>\<sigma>)`
    and `\<not> l \<rhd> u`
  show ?thesis by blast
qed

lemma SN_var:
 assumes "wwf_qtrs Q R" shows "SN_elt (qrstep Q R) (Var x)" (is "SN_elt _ ?x")
proof -
  let ?qr = "qrstep Q R"
  show "SN_elt ?qr ?x"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    then obtain A where "A 0 = ?x" and rsteps: "\<forall>i. (A i,A(Suc i)) \<in> ?qr"
      unfolding SN_elt_def by best
    hence "(?x,A 1) \<in> ?qr" using spec[OF rsteps, of 0] by auto
    then obtain l r C \<sigma> where "(l,r) \<in> R" and x: "Var x = C\<langle>l \<cdot> \<sigma>\<rangle>" by auto
    with wwf_qtrs_imp_left_fun and assms obtain f ls where "l = Fun f ls" by blast
    with x obtain ls where "Var x = C\<langle>Fun f ls\<rangle>" by auto
    thus False by (cases C, auto)
  qed
qed

lemma Tinf_var:
 assumes "wwf_qtrs Q R" shows "\<not> Tinf (qrstep Q R) (Var x)"
using SN_var[OF assms] unfolding Tinf_def by auto
 
lemma Tinf_lemma:
 assumes wwf: "wwf_qtrs Q R" and tinf: "Tinf (qrstep Q R) t"
 shows "\<exists>l r u \<sigma>. (l,r) \<in> useful_rules Q R \<and> r \<unrhd> u \<and> (\<forall>x. u \<noteq> Var x) \<and> (t,l\<cdot>\<sigma>) \<in> (nrqrstep Q R)^* \<and> (l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> rqrstep Q R
        \<and> r\<cdot>\<sigma> \<unrhd> (u\<cdot>\<sigma>) \<and> Tinf (qrstep Q R) (l\<cdot>\<sigma>) \<and> Tinf (qrstep Q R) (u\<cdot>\<sigma>) \<and> \<not>(l \<rhd> u)"
proof (cases t)
  case (Var x)
  with Tinf_var[OF wwf] tinf have False by auto
  thus ?thesis ..
next
  case (Fun f ts)
  from Tinf_lemma_1[OF wwf tinf[simplified Fun]]
  show ?thesis unfolding Fun .
qed

lemma Tinf_imp_def_root_qrstep:
  assumes wwf: "wwf_qtrs Q R" and tinf: "Tinf (qrstep Q R) t"
  shows "\<exists>f ss. t = Fun f ss \<and> defined (useful_rules Q R) (f, length ss)"
proof -
  from Tinf_var[OF wwf] tinf obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  with tinf have tinf: "Tinf (qrstep Q R) (Fun f ts)" by auto
  from Tinf_lemma[OF wwf tinf]
  obtain l r \<sigma> where lr: "(l,r) \<in> useful_rules Q R" and "(Fun f ts,l\<cdot>\<sigma>) \<in> (nrqrstep Q R)^*" 
    by blast
  with wwf[unfolded wwf_qtrs_wf_trs wf_trs_def]
  obtain g ls where l: "l = Fun g ls" by blast
  with nrqrsteps_preserve_root[OF `(Fun f ts,l\<cdot>\<sigma>) \<in> (nrqrstep Q R)^*`]
    nrqrsteps_num_args[OF `(Fun f ts,l\<cdot>\<sigma>) \<in> (nrqrstep Q R)^*`]
  have "g = f" and "length ls = length ts" by auto
  with lr l have "defined (useful_rules Q R) (f, length ts)" unfolding defined_def by auto
  with t show ?thesis by auto
qed

definition
  wf_qtrs :: "('f, 'v) term set \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool"
where
  "wf_qtrs Q R \<equiv> wwf_qtrs Q R \<and> (\<forall>t\<in>Q. \<not> is_var t)"

lemma nrqrstep_imp_sharp_qrstep:
  assumes step: "(s,t) \<in> nrqrstep Q R"
  shows "(sharp_term shp s, sharp_term shp t) \<in> qrstep Q R"
proof -
  let ?qr = "qrstep Q R"
  let ?nr = "nrqrstep Q R"
  let ?Q = "Q"
  let ?R = "R"
  let ?Iqr = "qrstep ?Q ?R"
  from step obtain l r C \<sigma> where C: "C \<noteq> \<box>" and  lr: "(l,r) \<in> R" and s: "s = C\<langle>l\<cdot>\<sigma>\<rangle>" and t: "t = C\<langle>r\<cdot>\<sigma>\<rangle>" and NF: "\<forall>u\<lhd>l \<cdot> \<sigma>.  u \<in> NF_terms Q" 
    unfolding nrqrstep_def rstep_r_c_s_def by auto
  from C obtain D f bef aft where C: "C = More f bef D aft" by (cases C, auto)
  let ?Dl = "D\<langle>l \<cdot> \<sigma>\<rangle>"
  let ?Dr = "D\<langle>r \<cdot> \<sigma>\<rangle>"
  let ?C = "More (shp f) bef \<box> aft"
  from s t C have s: "s = Fun f (bef @ ?Dl # aft)" and t: "t = Fun f (bef @ ?Dr # aft)" by auto
  from lr NF have "(?Dl,?Dr) \<in> ?qr" by auto
  hence "(?C\<langle>?Dl\<rangle>,?C\<langle>?Dr\<rangle>) \<in> ?Iqr" by (rule qrstep_closed_ctxt )
  with s t show ?thesis by auto
qed

lemma nrqrsteps_imp_sharp_qrsteps:
  assumes step: "(s,t) \<in> (nrqrstep Q R)^*"
  shows "(sharp_term shp s, sharp_term shp t) \<in> (qrstep Q R)^*"
using step
proof (induct rule: rtrancl_induct, simp)
  case (step u v)
  from nrqrstep_imp_sharp_qrstep[OF step(2), of shp] step(3)
  show ?case by auto
qed

lemma not_NF_termsE[elim]:
  assumes "s \<notin> NF_terms Q"
  obtains l C \<sigma> where "l \<in> Q" and "s = C\<langle>l \<cdot> \<sigma>\<rangle>"
proof -
  from assms obtain t where "(s, t) \<in> rstep (Id_on Q)" by auto
  with `\<And>l C \<sigma>. \<lbrakk>l \<in> Q; s = C\<langle>l \<cdot> \<sigma>\<rangle>\<rbrakk> \<Longrightarrow> thesis` show ?thesis by auto
qed

lemma Tinf_lemma_DP_explicit_subst:
  assumes wf: "wf_qtrs Q R" and s: "Tinf (qrstep Q R) s" and ndef: "\<And> f n ss. defined (useful_rules Q R) (f,n) \<Longrightarrow> length ss = n \<Longrightarrow> Fun (shp f) ss \<notin> Q"
  shows "\<exists>l r u \<sigma>. (l,r) \<in> DP shp (useful_rules Q R) \<and> r \<cdot> \<sigma> = sharp_term shp u
    \<and> Tinf (qrstep Q R) u \<and> (sharp_term shp s, l \<cdot> \<sigma>) \<in> (qrstep Q R)^*
    \<and> l \<cdot> \<sigma> \<in> NF_terms Q"
proof -
  let ?qr = "qrstep Q R"
  let ?U = "useful_rules Q R"
  let ?nrqr = "nrqrstep Q R"
  let ?rqr = "rqrstep Q R"
  from wf have wf: "wwf_qtrs Q R" and wfQ: "\<forall>t\<in>Q. \<not> is_var t"
    unfolding wf_qtrs_def by auto
  let ?Q = "Q" and ?R = "R"
  let ?Iqr = "qrstep ?Q ?R"
  obtain l r u \<sigma> where lr: "(l,r) \<in> ?U" and "r \<unrhd> u" and "\<forall>x. u \<noteq> Var x" and "(s,l\<cdot>\<sigma>) \<in> ?nrqr^*"
    and step: "(l\<cdot>\<sigma>,r\<cdot>\<sigma>) \<in> ?rqr" and "Tinf ?qr (l\<cdot>\<sigma>)" and "Tinf ?qr (u\<cdot>\<sigma>)" and "\<not> l \<rhd> u"
    using Tinf_lemma[OF wf s] by auto
  let ?u = "u\<cdot>\<sigma>"
  let ?l = "sharp_term shp l"
  let ?r = "sharp_term shp u"
  let ?\<sigma> = \<sigma>
  obtain f us where "u = Fun f us" by (cases u) (auto simp: `\<forall>x. u \<noteq> Var x`)
  have "root u = root ?u" unfolding `u = Fun f us` by simp
  obtain h vs where "?u = Fun h vs" and "defined ?U (h, length vs)" using Tinf_imp_def_root_qrstep[OF wf `Tinf ?qr ?u`] by auto
  from `(s,l\<cdot>\<sigma>) \<in> ?nrqr^*` have "(sharp_term shp s, sharp_term shp (l \<cdot> \<sigma>)) \<in> ?Iqr^*"
    by (rule nrqrsteps_imp_sharp_qrsteps)
  from `(l,r) \<in> ?U` and wf[unfolded wwf_qtrs_wf_trs] obtain g ts where l: "l = Fun g ts" unfolding wf_trs_def by blast
  have "sharp_term shp (l\<cdot>\<sigma>) = ?l\<cdot>?\<sigma>" unfolding `l = Fun g ts` sharp_term_subst_distr by simp
  with `(sharp_term shp s,sharp_term shp (l\<cdot>\<sigma>)) \<in> ?Iqr^*` have "(sharp_term shp s,?l\<cdot>?\<sigma>) \<in> ?Iqr^*" by simp
  from `(u\<cdot>\<sigma>) = Fun h vs` and `u = Fun f us` have "f = h" and "length us = length vs" by auto
  with `defined ?U (h, length vs)` have "defined ?U (f, length us)" by simp
  from `(l,r) \<in> ?U` and `r \<unrhd> u` and `defined ?U (f, length us)` and `\<not>(l \<rhd> u)` have "(?l,?r) \<in> DP shp ?U" unfolding DP_def `u = Fun f us` by auto
  have "?r\<cdot>?\<sigma> = sharp_term shp ?u" unfolding `u = Fun f us` sharp_term_subst_distr by simp
  from step have NF: "\<forall>u\<lhd>l\<cdot>\<sigma>. u \<in> NF_terms Q" unfolding rqrstep_def rstep_r_c_s_def by auto
  from lr l have gdef: "defined ?U (g, length ts)" unfolding defined_def by auto
  note ndef = ndef[OF gdef]
  note NF = NF[unfolded l]
  have "?l \<cdot> ?\<sigma> \<in> NF_terms ?Q"
  proof (rule ccontr)
    assume "\<not> ?thesis"
    from not_NF_termsE[OF this]
    obtain q C \<gamma> where q: "q \<in> ?Q" and match: "?l \<cdot> ?\<sigma> = C\<langle>q \<cdot> \<gamma>\<rangle>" .
    from q have qg: "q \<cdot> \<gamma> \<notin> NF_terms Q" unfolding NF_def by force
    show False
    proof (cases C)
      case (More ff bef D aft)
      hence "Fun g ts \<cdot> ?\<sigma> \<rhd> q \<cdot> \<gamma>" using match l by auto
      with NF have "q \<cdot> \<gamma> \<in> NF_terms Q" by auto
      with qg show False by simp
    next
      case Hole
      with match have match: "?l\<cdot>?\<sigma> = q \<cdot> \<gamma>" by auto
      show ?thesis 
      proof (cases q)
        case (Fun ff qq)
        from match[unfolded l this] have ff: "ff = shp g" and qq: "length ts = length qq" using map_eq_imp_length_eq[of _ ts _ qq] by auto
        from ndef[OF qq[symmetric]] q[unfolded Fun ff]  
        show ?thesis by auto
      next
        case (Var x)
        from q[unfolded Var] wfQ show False by auto
      qed
    qed 
  qed
  hence "?l \<cdot> ?\<sigma> \<in> NF_terms ?Q" by auto
  with `(?l,?r) \<in> DP shp ?U` and `?r\<cdot>?\<sigma> = sharp_term shp ?u` and `Tinf ?qr ?u` and `(sharp_term shp s,?l\<cdot>?\<sigma>) \<in> ?Iqr^*`
  show ?thesis by blast
qed


lemma SN_args_imp_SN :
  assumes SN: "\<And> s. s \<in> set ss \<Longrightarrow> SN_elt (qrstep Q R) s"
  and wwf: "wwf_qtrs Q R"
  and ndef: "\<not> defined (useful_rules Q R) (f,length ss)"
  shows "SN_elt (qrstep Q R) (Fun f ss)"
proof (rule ccontr)
  assume nSN: "\<not> SN_elt (qrstep Q R) (Fun f ss)"
  have SN: "\<And> s. Fun f ss \<rhd> s \<Longrightarrow> SN_elt (qrstep Q R) s"
  proof -
    fix s
    assume "Fun f ss \<rhd> s"
    then obtain C where "C \<noteq> \<box>" and "Fun f ss = C\<langle>s\<rangle>" ..
    then obtain bef D aft where "ss = bef @ D\<langle>s\<rangle> # aft" by (cases C, auto)
    hence "D\<langle>s\<rangle> \<in> set ss" by simp
    from SN[OF this] not_SN_subt_imp_not_SN[OF ctxt_closed_qrstep _ ctxt_imp_supteq[of D s]]
    show "SN_elt (qrstep Q R) s" by auto
  qed
  let ?t = "Fun f ss"
  let ?R = "useful_rules Q R"
  have t: "Tinf (qrstep Q R) ?t" unfolding Tinf_def using nSN SN by best
  from Tinf_lemma[OF wwf t] obtain l r \<sigma> where lr: "(l,r) \<in> useful_rules Q R" and
    steps: "(?t, l \<cdot> \<sigma>) \<in> (nrqrstep Q R)^*" and "(l \<cdot> \<sigma>, r \<cdot> \<sigma>) \<in> rqrstep Q R" by auto
  from nrqrsteps_preserve_root[OF steps] obtain ts where ls: "l \<cdot> \<sigma> = Fun f ts" by auto
  from nrqrsteps_num_args[OF steps, unfolded ls] have len: "length ts = length ss" by auto
  from wwf[unfolded wwf_qtrs_def] lr[unfolded useful_rules_def] have "\<not> is_var l" by auto
  with ls len obtain ls where l: "l = Fun f ls" and len: "length ss = length ls" by (cases l, auto)
  hence "defined ?R (f, length ss)" unfolding defined_def using lr by auto
  with ndef show False .. 
qed

lemma someI_ex4:
  "\<exists>w x y z. P (w, x, y, z) \<Longrightarrow> P (SOME (w, x, y, z). P (w, x, y, z))"
by (simp add: some_eq_ex)

fun
  next4 :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f,'v)term set \<Rightarrow> ('f,'v)trs \<Rightarrow> ('f,'v)term \<Rightarrow>
    ('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)term \<times> ('f,'v)subst"
where
  "next4 shp Q R s = (SOME (l, r, u, \<sigma>).
    (l, r) \<in> DP shp (useful_rules Q R) \<and> r \<cdot> \<sigma> = sharp_term shp u \<and> Tinf (qrstep Q R) u
    \<and> (sharp_term shp s, l \<cdot> \<sigma>) \<in> (qrstep Q R)^*
    \<and> l \<cdot> \<sigma> \<in> NF_terms Q)"

fun
  concrete_seq :: "('f \<Rightarrow> 'f) \<Rightarrow> ('f, 'v) term set \<Rightarrow> ('f, 'v) trs \<Rightarrow> ('f, 'v) term \<Rightarrow> nat \<Rightarrow>
    ('f, 'v) term \<times> ('f, 'v) term \<times> ('f, 'v) term \<times> ('f, 'v) term \<times> ('f, 'v) subst"
where
  "concrete_seq shp Q R s 0 = (s, next4 shp Q R s)" |
  "concrete_seq shp Q R r (Suc n) = (
    let ( _, _, _, s, _) = concrete_seq shp Q R r n in
    (s, next4 shp Q R s))"



theorem not_SN_imp_min_ichain:
  assumes "wf_qtrs Q R" and "\<not> SN (qrstep Q R)" 
  and ndef: "\<And> f n ss. defined (useful_rules Q R) (f,n) \<Longrightarrow> length ss = n \<Longrightarrow> Fun (shp f) ss \<notin> Q"
  and ndef2: "\<And> f n. defined (useful_rules Q R) (f,n) \<Longrightarrow> \<not> defined (useful_rules Q R) (shp f,n)"
  shows "\<exists>s t \<sigma>. min_ichain (initial_dpp shp Q R) s t \<sigma>"
proof -
  let ?qr = "qrstep Q R"
  let ?Q = "Q" and ?R = "R"
  let ?Iqr = "qrstep ?Q ?R"
  from `\<not>(SN(qrstep Q R))` obtain s' where "\<not> (SN_elt ?qr s')"
    unfolding SN_def by auto
  then obtain s where "s' \<unrhd>  s" and "Tinf ?qr s" using not_SN_imp_subt_Tinf by auto
  from assms have "wf_qtrs Q R" by auto
  let ?I = "\<lambda>i. fst(concrete_seq shp Q R s i)"
  let ?S = "\<lambda>i. fst(snd(concrete_seq shp Q R s i))"
  let ?T = "\<lambda>i. fst(snd(snd(concrete_seq shp Q R s i)))"
  let ?\<sigma> = "\<lambda>i. snd(snd(snd(snd(concrete_seq shp Q R s i))))"
  let ?P = "\<lambda>s l r u \<sigma>. (l,r) \<in> DP shp (useful_rules Q R) \<and> r\<cdot>\<sigma> = sharp_term shp u
    \<and> Tinf ?qr u \<and> (sharp_term shp s,l\<cdot>\<sigma>) \<in> ?Iqr^* \<and> l \<cdot> \<sigma> \<in> NF_terms ?Q"
  let ?P' =  "\<lambda>(l,r,u,\<sigma>). ?P s l r u \<sigma>"
  have main: "\<forall>i. ?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
  proof
    fix i show "?P (?I i) (?S i) (?T i) (?I(Suc i)) (?\<sigma> i)"
    proof (induct i)
      case 0 
      have "\<exists>l r u \<sigma>. ?P'(l,r,u,\<sigma>)"
        using Tinf_lemma_DP_explicit_subst[OF `wf_qtrs Q R` `Tinf ?qr s` ndef] by auto
      hence "?P'(next4 shp Q R s)" unfolding next4.simps using someI_ex4[where P = "?P'"] by best
      have ini: "?I 0 = s" unfolding concrete_seq.simps by simp
      have s: "?S 0 = fst(next4 shp Q R s)" unfolding concrete_seq.simps by simp
      have t: "?T 0 = fst(snd(next4 shp Q R s))" unfolding concrete_seq.simps by simp
      have sub: "?\<sigma> 0 = snd(snd(snd(next4 shp Q R s)))"
        unfolding concrete_seq.simps Let_def split_def by auto
      have nxt: "?I(Suc 0) = fst(snd(snd(next4 shp Q R s)))"
        unfolding concrete_seq.simps Let_def split_def by auto
      show ?case unfolding ini s t sub nxt using `?P'(next4 shp Q R s)` by auto
    next
      case (Suc i)
      let ?P'' = "\<lambda>(w,x,y,z). ?P (?I(Suc i)) w x y z"
      from Suc have "Tinf ?qr (?I(Suc i))" by simp
      have "\<exists>l r u \<sigma>. ?P''(l,r,u,\<sigma>)"
        using Tinf_lemma_DP_explicit_subst[OF `wf_qtrs Q R` `Tinf ?qr (?I(Suc i))` ndef] by auto
      hence IH: "?P''(next4 shp Q R (?I(Suc i)))"
        unfolding next4.simps using someI_ex4[where P = "?P''"] by simp
      have s: "?S(Suc i) = fst(next4 shp Q R (?I(Suc i)))"
        unfolding concrete_seq.simps Let_def split_def by simp
      have t: "?T(Suc i) = fst(snd(next4 shp Q R (?I(Suc i))))"
        unfolding concrete_seq.simps Let_def split_def by simp
      have sub: "?\<sigma>(Suc i) = snd(snd(snd(next4 shp Q R (?I(Suc i)))))"
        unfolding concrete_seq.simps Let_def split_def by simp
      have nxt: "?I(Suc(Suc i)) = fst(snd(snd(next4 shp Q R (?I(Suc i)))))"
        unfolding concrete_seq.simps Let_def split_def by simp
      show ?case unfolding s t sub nxt using IH by auto
    qed
  qed
  from main have P: "\<forall>i. (?S i,?T i) \<in> DP shp (useful_rules Q R)" by simp
  from main have NF: "\<And>i. (?S i \<cdot> ?\<sigma> i) \<in> NF_terms ?Q" by auto
  from main have eq: "\<And>i. sharp_term shp (?I(Suc i)) = ?T i \<cdot> ?\<sigma> i" by auto
  have R: "\<forall>i. (?T i \<cdot> ?\<sigma> i,?S(Suc i) \<cdot> ?\<sigma>(Suc i)) \<in> ?Iqr^*"
  proof
    fix i show "(?T i \<cdot> ?\<sigma> i,?S(Suc i)\<cdot>?\<sigma>(Suc i)) \<in> ?Iqr^*"
    proof (cases i)
      case 0
      from main have first: "?T 0 \<cdot> ?\<sigma> 0 = sharp_term shp (?I(Suc 0))" by best
      from main have "(sharp_term shp (?I(Suc 0)),?S(Suc 0)\<cdot>?\<sigma>(Suc 0)) \<in> ?Iqr^*" by best
      thus ?thesis unfolding 0 unfolding first by simp
    next
      case (Suc i) show ?thesis unfolding eq[symmetric] using main by best
    qed
  qed
  from P and R and NF have ichain: "ichain (initial_dpp shp Q R) ?S ?T ?\<sigma>" by auto
  have Tinf: "\<forall>i. Tinf ?qr (?I i)"
  proof
    fix i show "Tinf ?qr (?I i)" using main by (induct i) (auto simp: `Tinf ?qr s`)
  qed
  have SN_t: "\<forall>i. SN_elt ?Iqr (?T i \<cdot> ?\<sigma> i)"
  proof
    fix i
    from Tinf have Tinf: "Tinf ?qr (?I(Suc i))" by best
    from `wf_qtrs Q R` have wwf: "wwf_qtrs Q R" unfolding wf_qtrs_def by auto
    from main have "(?S i, ?T i) \<in> DP shp (useful_rules Q R)" by auto
    then obtain f ts where "?T i = Fun (shp f) ts" and "defined (useful_rules Q R) (f,length ts)" unfolding DP_def by auto
    then obtain ts where id: "?T i \<cdot> ?\<sigma> i = Fun (shp f) ts" and d: "defined (useful_rules Q R) (f,length ts)"  by auto
    from eq id have eq: "sharp_term shp (?I(Suc i)) = Fun (shp f) ts" by simp
    show "SN_elt ?Iqr (?T i \<cdot> ?\<sigma> i)" unfolding id 
    proof (rule SN_args_imp_SN[OF _ wwf ndef2[OF d]])
      fix t
      assume t: "t \<in> set ts"
      from eq obtain g where "?I (Suc i) = Fun g ts" by (cases "?I (Suc i)", auto)
      from Tinf[unfolded this] t show "SN_elt ?qr t" unfolding Tinf_def by auto
    qed
  qed
  from ichain and SN_t
  show ?thesis unfolding initial_dpp.simps min_ichain.simps by blast
qed


hide_const next4 concrete_seq



corollary not_SN_imp_min_ichain_rstep:
  assumes wf: "wf_trs R" and nSN: "\<not> SN (rstep R)" and ndef: "\<And> f n. defined R (f,n) \<Longrightarrow> \<not> defined R (shp f,n)"
  shows "\<exists> s t \<sigma>. min_ichain (DP shp R, {}, R) s t \<sigma>"
proof -
  let ?DP = "DP shp (useful_rules {} R)"
  have U: "useful_rules {} R = R" unfolding useful_rules_def useful_rule_def by auto
  from wf U have wf: "wf_qtrs {} R" unfolding wf_qtrs_def wwf_qtrs_wf_trs by auto
  from nSN have nSN: "\<not> SN (qrstep {} R)" unfolding rstep_qrstep_conv .
  have id: "initial_dpp shp {} R = (DP shp R, {}, R)" by (simp add: useful_rules_def useful_rule_def)
  from not_SN_imp_min_ichain[OF wf nSN, of shp, unfolded id U, OF _ ndef] show ?thesis by auto
qed

end
