(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)

header {* Term *}

theory Term
imports Abstract_Rewriting2 Position Show Util2
begin

text {* A theory on first-order terms. *}

text {*
A \emph{signature} is a set of function symbols arity pairs,
where the arity of a function symbol, denotes the number of arguments it expects.
*}
type_synonym
  'f sig = "('f \<times> nat) set"

text {* (First-order) \emph{terms} are built from function symbols and variables. *}
datatype ('f, 'v) "term" = Var 'v | Fun 'f "('f, 'v) term list"

text {* Type abbreviation for infinite sequences of terms. *}
type_synonym ('f, 'v) itseq = "('f, 'v)term iseq"

text {* Resolve problems with recurisve dependencies in generated code. *}
lemma [code]:
  shows "HOL.eq (Var x) (Var y) \<longleftrightarrow> HOL.eq x y"
  and "HOL.eq (Fun f ts) (Fun g us) \<longleftrightarrow> HOL.eq f g \<and> list_all2 HOL.eq ts us"
  by (simp_all add: list_all2_eq [symmetric])

text {* An alternative induction scheme for terms. *}
lemma
 fixes P :: "('f,'v)term \<Rightarrow> bool"
 assumes "\<And>x. P(Var x)" and "\<And>f ss. (\<And>s. s \<in> set ss \<Longrightarrow> P s) \<Longrightarrow> P(Fun f ss)"
 shows term_induct[case_names Var Fun, induct type: "term"]: "P t" and "\<And>t. t \<in> set ts \<Longrightarrow> P t"
using assms by (induct t and ts) auto


subsection {* @{text "show"} Instance for Terms *}

instantiation "term" :: ("show", "show") "show" begin
  fun
    shows_term :: "('f::show, 'v::show) term \<Rightarrow> shows"
  where
    "shows_term (Var x) = shows x" |
    "shows_term (Fun f ts) = (
      shows f +@+ shows_list_gen id [] ''('' '', '' '')'' (map shows_term ts)
    )"
  definition "shows_prec (d::nat) t = shows_term t"
  definition "shows_list_term (is::('f::show, 'v::show) term list) = shows_list_aux shows is"

  lemma assoc_term: "shows_prec d (t::('f::show, 'v::show) term) r @ s = shows_prec d t (r @ s)"
  proof (induct t arbitrary: r s)
    case (Var x) show ?case unfolding shows_prec_term_def by simp
  next
    case (Fun f ts)
    hence "\<And>r s. \<forall>t\<in>set ts. id (shows_prec d t) r @ s = id (shows_prec d t) (r @ s)" by auto
    hence elt: "\<And>r s. \<forall>t\<in>set (map shows_term ts). id t r @ s = id t (r @ s)" unfolding shows_prec_term_def by auto
    show ?case unfolding shows_prec_term_def shows_term.simps o_def
      unfolding assoc
      unfolding shows_list_gen_assoc[OF elt,of "[]" "''(''" "'', ''" "'')''" r s] using Fun by simp
  qed
  lemma assoc_term_elt: "\<forall>(t::('f::show, 'v::show) term)\<in>set ts. shows t r @ s = shows t (r @ s)"
    using assoc_term by auto
 instance by
    (intro_classes,rule assoc_term,unfold shows_list_term_def)
    (rule shows_list_aux_assoc[OF assoc_term_elt])
end


subsection {* Variables *}

primrec is_var :: "('f, 'v) term \<Rightarrow> bool" 
where
  "is_var (Var x) = True" | 
  "is_var (Fun f ts) = False"

lemma is_varE[elim]: "is_var t \<Longrightarrow> (\<And>x. t = Var x \<Longrightarrow> P) \<Longrightarrow> P" by (cases t) simp_all
lemma is_var_Fun[elim]:
  assumes "\<not> is_var t" obtains f ts where "t = Fun f ts"
using assms by (cases t) auto

fun
  the_var :: "('f, 'v) term \<Rightarrow> 'v"
where
  "the_var (Var x) = x"

text {* The set of all variables occuring in a term. *}
fun
  vars_term :: "('f, 'v) term \<Rightarrow> 'v set"
where
  "vars_term (Var x) = {x}" |
  "vars_term (Fun f ts) = \<Union>set (map vars_term ts)"

text {* The list of variables in a term (without removing duplicates). *}
fun
  vars_term_list' :: "('f, 'v) term \<Rightarrow> 'v list"
where
  "vars_term_list' (Var x) = [x]" |
  "vars_term_list' (Fun f ts) = flat_map vars_term_list' ts"

lemma vars_term_list'_sound[simp]: "set (vars_term_list' t) = vars_term t"
by (induct t) simp_all
declare vars_term_list'.simps[simp del]

fun
  vars_term_list :: "('f, 'v) term \<Rightarrow> 'v list"
where
  "vars_term_list t = remdups (vars_term_list' t)"

lemma vars_term_list_sound[simp]: "set (vars_term_list t) = vars_term t" by simp
declare vars_term_list.simps[simp del]


text {* The set of all function symbols occuring in a term. *}
fun
  funs_term :: "('f, 'v) term \<Rightarrow> 'f set"
where
  "funs_term (Var x) = {}" |
  "funs_term (Fun f ts) = {f} \<union> \<Union>set (map funs_term ts)"

text {* The list of function symbols in a term (without removing duplicates). *}
fun
  funs_term_list' :: "('f, 'v) term \<Rightarrow> 'f list"
where
  "funs_term_list' (Var x) = []" |
  "funs_term_list' (Fun f ts) = f # flat_map funs_term_list' ts"

lemma funs_term_list'_sound[simp]: "set (funs_term_list' t) = funs_term t"
by (induct t) simp_all
declare funs_term_list'.simps[simp del]

fun
  funs_term_list :: "('f, 'v) term \<Rightarrow> 'f list"
where
  "funs_term_list t = remdups (funs_term_list' t)"

lemma funs_term_list_sound[simp]: "set (funs_term_list t) = funs_term t" by simp
declare funs_term_list.simps[simp del]

text {* The set of all function symbol and arity pairs occuring in a term. *}
fun
  funas_term :: "('f, 'v) term \<Rightarrow> 'f sig"
where
  "funas_term (Var x) = {}" |
  "funas_term (Fun f ts) = {(f, length ts)} \<union> \<Union>set (map funas_term ts)"

text {* The list of function symbol and arity pairs in a term (without removing duplicates). *}
fun
  funas_term_list' :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) list"
where
  "funas_term_list' (Var x) = []" |
  "funas_term_list' (Fun f ts) = (f, length ts) # flat_map funas_term_list' ts"

lemma funas_term_list'_sound[simp]: "set (funas_term_list' t) = funas_term t"
by (induct t) simp_all
declare funas_term_list'.simps[simp del]

fun
  funas_term_list :: "('f, 'v) term \<Rightarrow> ('f \<times> nat) list"
where
  "funas_term_list t = remdups (funas_term_list' t)"

lemma funas_term_list_sound[simp]: "set (funas_term_list t) = funas_term t" by simp
declare funas_term_list.simps[simp del]

text {* A term is called \emph{ground} if it does not contain
any variables. *}
fun
  ground :: "('f, 'v) term \<Rightarrow> bool"
where
  "ground (Var x) = False" |
  "ground (Fun f ts) = (\<forall>t\<in>set ts. ground t)"

text {* The \emph{root symbol} of a term is defined by: *}
fun
  root :: "('f, 'v) term \<Rightarrow> 'f option"
where
  "root (Var x) = None" |
  "root (Fun f ts) = Some f"

text {* The \emph{depth} of a term is defined as follows: *}
fun
  depth :: "('f, 'v) term \<Rightarrow> nat"
where
  "depth (Var x) = 0" |
  "depth (Fun f []) = 0" |
  "depth (Fun f ts) = 1 + Max (set (map depth ts))"


subsection {* Subterms *}

text {* The \emph{superterm} relation. *}
inductive_set
  supteq :: "(('f, 'v) term \<times> ('f, 'v) term) set"
where
  refl: "(t, t) \<in> supteq" |
  subt: "u \<in> set ss \<Longrightarrow> (u, t) \<in> supteq \<Longrightarrow> (Fun f ss, t) \<in> supteq"

text {* The \emph{proper superterm} relation. *}
inductive_set
  supt :: "(('f, 'v) term \<times> ('f, 'v) term) set"
where
  arg: "s \<in> set ss \<Longrightarrow> (Fun f ss, s) \<in> supt" |
  subt: "s \<in> set ss \<Longrightarrow> (s, t) \<in> supt \<Longrightarrow> (Fun f ss, t) \<in> supt"

hide_const suptp supteqp
hide_fact
  suptp.arg suptp.cases suptp.induct suptp.intros suptp.subt suptp_supt_eq
hide_fact
  supteqp.cases supteqp.induct supteqp.intros supteqp.refl supteqp.subt supteqp_supteq_eq

hide_fact (open) supt.arg supt.subt supteq.refl supteq.subt

subsubsection {* Syntax for Subterms *}

text {* Infix syntax. *}
abbreviation supt_pred where "supt_pred s t \<equiv> (s, t) \<in> supt"
abbreviation supteq_pred where "supteq_pred s t \<equiv> (s, t) \<in> supteq"
abbreviation (input) subt where "subt s t \<equiv> supt_pred t s"
abbreviation (input) subteq where "subteq s t \<equiv> supteq_pred t s"

notation
  supt ("op |>") and
  supt_pred ("(_/ |> _)" [56, 56] 55) and
  subt (infix "<|" 55) and
  supteq ("op |>=") and
  supteq_pred ("(_/ |>= _)" [56, 56] 55) and
  subteq (infix "<|=" 55)

notation (xsymbols)
  supt ("op \<rhd>") and
  supt_pred ("(_/ \<rhd> _)" [56, 56] 55) and
  subt (infix "\<lhd>" 55) and
  supteq ("op \<unrhd>") and
  supteq_pred ("(_/ \<unrhd> _)" [56, 56] 55) and
  subteq (infix "\<unlhd>" 55)

text {* Quantifier syntax. *}

syntax
  "_All_supteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3ALL _|>=_./ _)" [0, 0, 10] 10)
  "_Ex_supteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3EX _|>=_./ _)" [0, 0, 10] 10)
  "_All_supt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3ALL _|>_./ _)" [0, 0, 10] 10)
  "_Ex_supt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3EX _|>_./ _)" [0, 0, 10] 10)

  "_All_subteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3ALL _<|=_./ _)" [0, 0, 10] 10)
  "_Ex_subteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3EX _<|=_./ _)" [0, 0, 10] 10)
  "_All_subt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3ALL _<|_./ _)" [0, 0, 10] 10)
  "_Ex_subt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3EX _<|_./ _)" [0, 0, 10] 10)

syntax (xsymbols)
  "_All_supteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<forall>_\<unrhd>_./ _)" [0, 0, 10] 10)
  "_Ex_supteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<exists>_\<unrhd>_./ _)" [0, 0, 10] 10)
  "_All_supt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<forall>_\<rhd>_./ _)" [0, 0, 10] 10)
  "_Ex_supt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<exists>_\<rhd>_./ _)" [0, 0, 10] 10)

  "_All_subteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<forall>_\<unlhd>_./ _)" [0, 0, 10] 10)
  "_Ex_subteq" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<exists>_\<unlhd>_./ _)" [0, 0, 10] 10)
  "_All_subt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<forall>_\<lhd>_./ _)" [0, 0, 10] 10)
  "_Ex_subt" :: "[idt, 'a, bool] \<Rightarrow> bool" ("(3\<exists>_\<lhd>_./ _)" [0, 0, 10] 10)

(* for parsing *)
translations
  "ALL x|>=y. P" => "ALL x. x |>= y \<longrightarrow> P"
  "EX x|>=y. P"  => "EX x. x |>= y \<and> P"
  "ALL x|>y. P"  => "ALL x. x |> y \<longrightarrow> P"
  "EX x|>y. P"   => "EX x. x |> y \<and> P"
  
  "ALL x<|=y. P" => "ALL x. x <|= y \<longrightarrow> P"
  "EX x<|=y. P"  => "EX x. x <|= y \<and> P"
  "ALL x<|y. P"  => "ALL x. x <| y \<longrightarrow> P"
  "EX x<|y. P"   => "EX x. x <| y \<and> P"

(* for printing *)
print_translation {*
let
  val All_binder = Syntax.binder_name @{const_syntax All};
  val Ex_binder = Syntax.binder_name @{const_syntax Ex};
  val impl = @{const_syntax "implies"};
  val conj = @{const_syntax "conj"};
  val supt = @{const_syntax "supt_pred"};
  val supteq = @{const_syntax "supteq_pred"};

  val trans =
   [((All_binder, impl, supt), ("_All_supt", "_All_subt")),
    ((All_binder, impl, supteq), ("_All_supteq", "_All_subteq")),
    ((Ex_binder, conj, supt), ("_Ex_supt", "_Ex_subt")),
    ((Ex_binder, conj, supteq), ("_Ex_supteq", "_Ex_subteq"))];

  fun matches_bound v t = 
     case t of (Const ("_bound", _) $ Free (v', _)) => (v = v')
              | _ => false
  fun contains_var v = Term.exists_subterm (fn Free (x, _) => x = v | _ => false)
  fun mk v c n P = Syntax.const c $ Syntax.mark_bound v $ n $ P

  fun tr' q = (q,
    fn [Const ("_bound", _) $ Free (v, _), Const (c, _) $ (Const (d, _) $ t $ u) $ P] =>
      (case AList.lookup (op =) trans (q, c, d) of
        NONE => raise Match
      | SOME (l, g) =>
          if matches_bound v t andalso not (contains_var v u) then mk v l u P
          else if matches_bound v u andalso not (contains_var v t) then mk v g t P
          else raise Match)
     | _ => raise Match);
in [tr' All_binder, tr' Ex_binder] end
*}


subsubsection {* Transitivity Reasoning for Subterms *}

lemma supt_trans[trans]: "s \<rhd> t \<Longrightarrow> t \<rhd> u \<Longrightarrow> s \<rhd> u"
by (induct s t rule: supt.induct) (auto intro: supt.intros)

lemma supteq_trans[trans]: "s \<unrhd> t \<Longrightarrow> t \<unrhd> u \<Longrightarrow> s \<unrhd> u"
by (induct s t rule: supteq.induct) (auto intro: supteq.intros)

text {* Auxiliary lemmas about term size. *}
lemma size_simp5:
  "s \<in> set ss \<Longrightarrow> s \<rhd> t \<Longrightarrow> size t < size s \<Longrightarrow> size t < Suc (list_size size ss)"
by (induct ss) auto
lemma size_simp6:
  "s \<in> set ss \<Longrightarrow> (s, t) \<in> supt \<Longrightarrow> size t < size s \<Longrightarrow> size t < Suc (list_size size ss)"
by (induct ss) auto
lemma size_simp7:
  "s \<in> set ss \<Longrightarrow> s \<unrhd> t \<Longrightarrow> size t \<le> size s \<Longrightarrow> size t \<le> Suc (list_size size ss)"
by (induct ss) auto
lemma size_simp8:
  "s \<in> set ss \<Longrightarrow> (s, t) \<in> supteq \<Longrightarrow> size t \<le> size s \<Longrightarrow> size t \<le> Suc (list_size size ss)"
by (induct ss) auto
lemmas size_simps = size_simp1 size_simp2 size_simp3 size_simp4 size_simp5 size_simp6 size_simp7 size_simp8

lemma supt_size: "s \<rhd> t \<Longrightarrow> size s > size t"
by (induct rule: supt.induct) (auto simp: size_simps)

lemma supteq_size: "s \<unrhd> t \<Longrightarrow> size s \<ge> size t"
by (induct rule: supteq.induct) (auto simp: size_simps)

text {* Reflexivity and Asymmetry. *}

declare supteq.refl[iff]

lemma reflcl_supteq[simp]: "supteq^= = supteq" by auto

lemma trancl_supteq[simp]: "supteq^+ = supteq"
by (rule trancl_id) (auto simp: trans_def intro: supteq_trans)

lemma rtrancl_supteq[simp]: "supteq^* = supteq"
unfolding trancl_reflcl[symmetric] by auto

lemma eq_supteq: "s = t \<Longrightarrow> s \<unrhd> t" by auto

lemma supt_neqD: "s \<rhd> t \<Longrightarrow> s \<noteq> t" using supt_size by auto

lemma supteq_Var[simp]: "x \<in> vars_term t \<Longrightarrow> t \<unrhd> Var x"
proof (induct t)
  case (Var y) thus ?case by (cases "x = y") auto
next
  case (Fun f ss) thus ?case by (auto intro: supteq.intros)
qed

lemma term_not_arg[iff]: "Fun f ss \<notin> set ss" (is "?t \<notin> set ss")
proof
  assume "?t \<in> set ss"
  hence "?t \<rhd> ?t" by (auto intro: supt.intros)
  hence "?t \<noteq> ?t" by (auto dest: supt_neqD)
  thus False by simp
qed

lemma supt_Fun[simp]: assumes "s \<rhd> Fun f ss" (is "s \<rhd> ?t") and "s \<in> set ss" shows "False"
proof -
  from `s \<in> set ss` have "?t \<rhd> s" by (auto intro: supt.intros)
  hence "size ?t > size s" by (rule supt_size)
  from `s \<rhd> ?t` have "size s > size ?t" by (rule supt_size)
  with `size ?t > size s` show False by simp
qed

lemma supt_supteq_conv: "s \<rhd> t = (s \<unrhd> t \<and> s \<noteq> t)"
proof
  assume "s \<rhd> t" thus "s \<unrhd> t \<and> s \<noteq> t"
  proof (induct rule: supt.induct)
    case (arg t ss f) show ?case
    proof
      from arg show "Fun f ss \<unrhd> t" by (auto intro: supteq.intros)
    next
      from arg have "size t < size (Fun f ss)" by (auto simp: size_simps)
      thus "Fun f ss \<noteq> t" by auto
    qed
  next
    case (subt s ss t f)
    have "s \<unrhd> s" ..
    from `s \<in> set ss` have "Fun f ss \<unrhd> s" by (auto intro: supteq.intros)
    from `s \<unrhd> t \<and> s \<noteq> t` have "s \<unrhd> t" ..
    with `Fun f ss \<unrhd> s` have first: "Fun f ss \<unrhd> t" by (rule supteq_trans)
    from `s \<in> set ss` and `s \<rhd> t` have "Fun f ss \<rhd> t" ..
    hence second: "Fun f ss \<noteq> t" by (auto dest: supt_neqD)
    from first and second show "Fun f ss \<unrhd> t \<and> Fun f ss \<noteq> t" by auto
  qed
next
  assume "s \<unrhd> t \<and> s \<noteq> t"
  hence "s \<unrhd> t" and "s \<noteq> t" by auto
  thus "s \<rhd> t" by (induct rule: supteq.induct) (auto intro: supt.intros)
qed

lemma supt_not_sym: "s \<rhd> t \<Longrightarrow> \<not> (t \<rhd> s)"
proof
 assume "s \<rhd> t" and "t \<rhd> s" hence "s \<rhd> s" by (rule supt_trans)
 thus False unfolding supt_supteq_conv by simp
qed

lemma supt_irrefl[iff]: "\<not> t \<rhd> t"
using supt_not_sym[of t t] by auto

lemma supt_imp_supteq: "s \<rhd> t \<Longrightarrow> s \<unrhd> t"
unfolding supt_supteq_conv by auto

lemma supt_supteq_not_supteq: "s \<rhd> t = (s \<unrhd> t \<and> \<not> (t \<unrhd> s))"
using supt_not_sym unfolding supt_supteq_conv by auto

lemma supteq_supt_conv: "(s \<unrhd> t) = (s \<rhd> t \<or> s = t)" by (auto simp: supt_supteq_conv)

lemma supteq_antisym:
  assumes "s \<unrhd> t" and "t \<unrhd> s" shows "s = t"
using assms unfolding supteq_supt_conv by (auto simp: supt_not_sym)


text {* The subterm relation is an order on terms. *}
interpretation subterm: order "op \<unlhd>" "op \<lhd>"
by (unfold_locales) (rule supt_supteq_not_supteq,
  auto intro: supteq_trans supteq_antisym supt_supteq_not_supteq)

lemma suptE[elim]: "s \<rhd> t \<Longrightarrow> (s \<unrhd> t \<Longrightarrow> P) \<Longrightarrow> (s \<noteq> t \<Longrightarrow> P) \<Longrightarrow> P"
by (auto simp: supt_supteq_conv)

lemma suptI[intro]: "s \<unrhd> t \<Longrightarrow> s \<noteq> t \<Longrightarrow> s \<rhd> t"
by (induct rule: supteq.induct) (auto intro: supt.intros)

lemma supt_supteq_set_conv: "op \<rhd> = op \<unrhd> \<inter> -Id" using supt_supteq_conv[to_set] by auto

lemma supteq_supt_set_conv: "supteq = supt \<union> Id"
by (auto simp: supt_supteq_conv intro: supteq.intros)

lemmas [intro] = supteq.intros supteq.subt

declare conj_cong[fundef_cong]
declare size_simps[simp]
fun
  poss :: "('f, 'v) term \<Rightarrow> pos set"
where
  "poss (Var x) = {\<epsilon>}" |
  "poss (Fun f ss) = {\<epsilon>} \<union> {i<#p | i p. i < length ss \<and> p \<in> poss (ss!i)}"
declare size_simps[simp del]
declare conj_cong[fundef_cong del]

fun
  poss_list :: "('f, 'v) term \<Rightarrow> pos list"
where
  "poss_list (Var x) = [\<epsilon>]" |
  "poss_list (Fun f ss) = (\<epsilon> # concat (map (\<lambda> (i,ps).
    map ( op <# i) ps) (zip [0 ..< length ss] (map poss_list ss))))"

lemma poss_list_sound[simp]: "set (poss_list s) = poss s" 
proof (induct s, simp)
  case (Fun f ss)
  let ?z = "zip [0..<length ss] (map poss_list ss)"
  have len: "length ?z = length ss" by auto
  have "(\<Union>a\<in>set ?z.
         set (split (\<lambda>i. map (op <# i)) a)) = 
       {i <# p |i p. i < length ss \<and> p \<in> poss (ss ! i)}" (is "?l = ?r")
  proof (rule set_eqI)
    fix ip
    show "(ip \<in> ?l) = (ip \<in> ?r)"
    proof
      assume "ip \<in> ?l" 
      from this obtain ipI where 
	z: "ipI \<in> set ?z" and 
	ip: "ip \<in> set (split (\<lambda> i. map (op <# i)) ipI)" 
	by auto     
      from z obtain i where i: "i < length ?z" and zi: "?z ! i = ipI"
        by (force simp: set_conv_nth)      
      with ip Fun show "ip \<in> ?r" by auto
    next
      assume "ip \<in> ?r"
      from this obtain i p where i: "i < length ss" and "p \<in> poss (ss ! i)"
	and ip: "ip = i <# p" by auto
      with Fun have p: "p \<in> set (poss_list (ss ! i))" and iz: "i < length ?z" by auto
      from i have id: "?z ! i = (i, poss_list (ss ! i))" (is "_ = ?ipI") by auto
      from iz have  "?z ! i \<in> set ?z" by (rule nth_mem)
      with id have inZ: "?ipI \<in> set ?z" by auto
      from p have "i <# p \<in> set (split (\<lambda> i. map (op <# i)) ?ipI)" by auto
      with inZ ip show "ip \<in> ?l" by force
    qed
  qed
  with Fun show ?case by simp
qed

declare poss_list.simps[simp del]

text {* The subterm relation is strongly normalizing. *}
lemma SN_supt: "SN (op \<rhd>)" unfolding SN_iff_wf
by (rule wf_subset) (auto intro: supt_size)

text {* The subterm of a term~@{text s} at position~@{text p} is defined as follows: *}
fun
  subt_at :: "('f, 'v) term \<Rightarrow> pos \<Rightarrow> ('f, 'v) term" ("(_ |'_ _)" [60])
where
  "s |_ \<epsilon> = s" |
  "Fun f ss |_ i<#p = ss!i|_p"

text {* A \emph{context} is a term containing exactly one \emph{hole}. *}
datatype ('f, 'v) ctxt =
  Hole ("\<box>") |
  More 'f "('f, 'v) term list" "('f, 'v) ctxt" "('f, 'v) term list"


subsection {* @{text "shows"} Instance for Contexts *}

instantiation ctxt :: ("show", "show") "show" begin
  fun shows_ctxt :: "('f::show, 'v::show) ctxt \<Rightarrow> shows"
  where
    "shows_ctxt(Hole) = shows_string ''[]''" |
    "shows_ctxt(More f ss1 D ss2) = (
       shows f +@+ ''('' +#+
       shows_list_gen shows [] [] '', '' '', '' ss1 +@+
       shows_ctxt D +@+
       shows_list_gen shows '')'' '', '' '', '' '')'' ss2
     )"
  definition "shows_prec (d::nat) C = shows_ctxt C"
  definition "shows_list_ctxt (cs::('f::show,'v::show) ctxt list) \<equiv> shows_list_aux shows cs"

  lemma assoc_ctxt:
    "shows_prec d (c::('f::show, 'v::show) ctxt) r @ s = shows_prec d c (r @ s)"
  by (induct c arbitrary: r s) (auto simp: shows_prec_ctxt_def)

  lemma assoc_ctxt_elt:
    "\<forall>(c::('f::show, 'v::show) ctxt)\<in>set cs. shows_prec d c r @ s = shows_prec d c (r @ s)"
  using assoc_ctxt by auto

  instance by
    (intro_classes,rule assoc_ctxt,unfold shows_list_ctxt_def)
    (rule shows_list_aux_assoc[OF assoc_ctxt_elt])
end

text {*
Replacing the hole in a context~@{text C} by a term ~@{text t} is called, application
of @{text C} to @{text t}.
*}
fun
  ctxt_apply :: "('f, 'v) ctxt \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term" ("_\<langle>_\<rangle>" [1000, 0] 1000)
where
  "\<box>\<langle>s\<rangle> = s" |
  "(More f ss1 C ss2)\<langle>s\<rangle> = Fun f (ss1 @ C\<langle>s\<rangle> # ss2)"

lemma ctxt_eq[simp]: "(C\<langle>s\<rangle> = C\<langle>t\<rangle>) = (s = t)" by (induct C) auto

fun
  ctxt_of_pos_term :: "pos \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) ctxt"
where
  "ctxt_of_pos_term \<epsilon> t = \<box>" |
  "ctxt_of_pos_term (i<#ps) (Fun f ts) =
    More f (take i ts) (ctxt_of_pos_term ps (ts!i)) (drop (Suc i) ts)"

lemma ctxt_supt_id:
  assumes "p \<in> poss t" shows "(ctxt_of_pos_term p t)\<langle>t|_p\<rangle> = t"
using assms by (induct t arbitrary: p) (auto simp: id_take_nth_drop[symmetric])

text {*
Let @{text s} and @{text t} be terms. The following three statements are equivalent:
\begin{enumerate}
 \item \label{A}@{text "s \<unrhd> t"}
 \item \label{B}@{text "\<exists>p\<in>poss s. s|_p = t"}
 \item \label{C}@{text "\<exists>C. s = C\<langle>t\<rangle>"}
\end{enumerate}
*}

fun
  ctxt_compose :: "('f, 'v) ctxt \<Rightarrow> ('f, 'v) ctxt \<Rightarrow> ('f, 'v) ctxt" (infixl "\<circ>c" 65)
where
  "\<box> \<circ>c D = D" |
  "(More f ss1 C ss2) \<circ>c D = More f ss1 (C \<circ>c D) ss2"

interpretation ctxt_monoid_mult: monoid_mult "\<box>" "op \<circ>c"
proof
  fix C D E show "C \<circ>c D \<circ>c E = C \<circ>c (D \<circ>c E)" by (induct C) simp_all
next
  fix C show "\<box> \<circ>c C = C" by simp
next
  fix C show "C \<circ>c \<box> = C" by (induct C) simp_all
qed

declare ctxt_monoid_mult.mult_assoc[simp]

instantiation ctxt :: (type, type) monoid_mult begin
  definition [simp]: "1 \<equiv> \<box>"
  definition [simp]: "op * = op \<circ>c"
  instance by (intro_classes) simp_all
end

lemma ctxt_compose_apply_distrib[simp]: "(C \<circ>c D)\<langle>t\<rangle> = C\<langle>D\<langle>t\<rangle>\<rangle>" by (induct C) simp_all

text {* The position of the hole in a context is uniquely determined. *}
fun
  hole_pos :: "('f, 'v) ctxt \<Rightarrow> pos"
where
  "hole_pos \<box> = \<epsilon>" |
  "hole_pos (More f ss D ts) = length ss <# hole_pos D"

lemma supteq_imp_subt_at:
  assumes "s \<unrhd> t" shows "\<exists>p\<in>poss s. s|_p = t"
using assms proof (induct s t rule: supteq.induct)
  case (refl s)
  have "\<epsilon> \<in> poss s" by (induct s rule: term_induct) auto
  have "s|_\<epsilon> = s" by simp
  from `\<epsilon> \<in> poss s` and `s|_\<epsilon> = s` show ?case by best
next
  case (subt u ss t f) 
  then obtain p where "p \<in> poss u" and "u|_p = t" by best
  from `u \<in> set ss` obtain i
    where "i < length ss" and "u = ss!i" by (auto simp: in_set_conv_nth)
  from `i < length ss` and `p \<in> poss u`
    have "i<#p \<in> poss (Fun f ss)" unfolding `u = ss!i` by simp
  have "Fun f ss|_i<#p = t"
    unfolding subt_at.simps unfolding `u = ss!i`[symmetric] by (rule `u|_p = t`)
  with `i<#p \<in> poss (Fun f ss)` show ?case by best
qed

lemma subt_at_imp_ctxt:
  assumes "p \<in> poss s" shows "\<exists>C. C\<langle>s|_p\<rangle> = s"
using assms proof (induct p arbitrary: s)
  case (Empty s)
  have "\<box>\<langle>s|_\<epsilon>\<rangle> = s" by simp
  thus ?case by best
next
  case (PCons i p s)
  then obtain f ss where "s = Fun f ss" by (cases s) auto
  with `i<#p \<in> poss s` obtain u :: "('a,'b) term"
    where "u = ss!i" and "p \<in> poss u" and "i < length ss" by auto
  from PCons and `p\<in>poss u` obtain D where "D\<langle>u|_p\<rangle> = u" by auto
  let ?ss1 = "take i ss" and ?ss2 = "drop (Suc i) ss"
  let ?E = "More f ?ss1 D ?ss2"
  have "?ss1@D\<langle>u|_p\<rangle>#?ss2 = ss" (is "?ss = ss") unfolding `D\<langle>u|_p\<rangle> = u` unfolding `u = ss!i`
    unfolding id_take_nth_drop[OF `i < length ss`, symmetric] ..
  have "s|_i<#p = u|_p" unfolding `s = Fun f ss` using `u = ss!i` by simp
  have "?E\<langle>s|_i<#p\<rangle> = s"
    unfolding ctxt_apply.simps `s|_i<#p = u|_p` `?ss = ss` unfolding `s = Fun f ss` ..
  thus ?case by best
qed

lemma ctxt_imp_supteq[simp]: shows "C\<langle>t\<rangle> \<unrhd> t" by (induct C) auto

lemma subt_at_imp_supteq': assumes "p \<in> poss s" and "s|_p = t" shows "s \<unrhd> t"
proof -
 from `p \<in> poss s` obtain C where "C\<langle>s|_p\<rangle> = s" using subt_at_imp_ctxt by best
 thus ?thesis unfolding `s|_p = t` using ctxt_imp_supteq by auto
qed

lemma subt_at_imp_supteq: assumes "p \<in> poss s" shows "s \<unrhd> s|_p"
proof -
 have "s|_p = s|_p" by auto
 with assms show ?thesis by (rule subt_at_imp_supteq')
qed

(* Conversions between contexts and subterms. *)
lemma supteq_ctxtE[elim]: assumes "s \<unrhd> t" obtains C where "s = C\<langle>t\<rangle>"
using assms proof (induct arbitrary: thesis)
  case (refl s)
  have "s = \<box>\<langle>s\<rangle>" by simp
  from refl[OF this] show ?case .
next
  case (subt u ss t f)
  then obtain C where "u = C\<langle>t\<rangle>" by auto
  from split_list[OF `u \<in> set ss`] obtain ss1 and ss2 where "ss = ss1 @ u # ss2" by auto
  hence "Fun f ss = (More f ss1 C ss2)\<langle>t\<rangle>" using `u = C\<langle>t\<rangle>` by simp
  with subt show ?case by best
qed

lemma ctxt_supteq[intro]: assumes "s = C\<langle>t\<rangle>" shows "s \<unrhd> t"
proof (cases C)
  case Hole thus ?thesis using assms by auto
next
  case (More f ss1 D ss2)
  with assms have s: "s = Fun f (ss1 @ D\<langle>t\<rangle> # ss2)" (is "_ = Fun _ ?ss") by simp
  have "D\<langle>t\<rangle> \<in> set ?ss" by simp
  moreover have "D\<langle>t\<rangle> \<unrhd> t" by (induct D) auto
  ultimately show ?thesis unfolding s ..
qed

lemma supteq_ctxt_conv: "(s \<unrhd> t) = (\<exists>C. s = C\<langle>t\<rangle>)" by auto

(* Conversions between contexts and proper subterms. *)
lemma supt_ctxtE[elim]: assumes "s \<rhd> t" obtains C where "C \<noteq> \<box>" and "s = C\<langle>t\<rangle>"
using assms proof (induct arbitrary: thesis)
  case (arg s ss f)
  from split_list[OF `s \<in> set ss`] obtain ss1 and ss2 where ss: "ss = ss1 @ s # ss2" by auto
  let ?C = "More f ss1 \<box> ss2"
  have "?C \<noteq> \<box>" by simp
  moreover have "Fun f ss = ?C\<langle>s\<rangle>" by (simp add: ss)
  ultimately show ?case using arg by best
next
  case (subt s ss t f)
  then obtain C where "C \<noteq> \<box>" and "s = C\<langle>t\<rangle>" by auto
  from split_list[OF `s \<in> set ss`] obtain ss1 and ss2 where ss: "ss = ss1 @ s # ss2" by auto
  have "More f ss1 C ss2 \<noteq> \<box>" by simp
  moreover have "Fun f ss = (More f ss1 C ss2)\<langle>t\<rangle>" using `s = C\<langle>t\<rangle>` by (simp add: ss)
  ultimately show ?case using subt(4) by best
qed

(* More transitivity rules. *)
lemma supt_supteq_trans[trans]:
  assumes "s \<rhd> t" and "t \<unrhd> u" shows "s \<rhd> u"
proof (cases "t = u")
  assume "t = u" with assms show ?thesis by simp
next
  assume "t \<noteq> u"
  with assms have "t \<rhd> u" by auto
  with `s \<rhd> t` show ?thesis by (rule supt_trans)
qed

lemma supteq_supt_trans[trans]:
  assumes "s \<unrhd> t" and "t \<rhd> u" shows "s \<rhd> u"
proof (cases "s = t")
  assume "s = t" with assms show ?thesis by simp
next
  assume "s \<noteq> t"
  with assms have "s \<rhd> t" by auto
  thus ?thesis using `t \<rhd> u` by (rule supt_trans)
qed

lemma ctxt_supt[intro 2]: assumes "C \<noteq> \<box>" and "s = C\<langle>t\<rangle>" shows "s \<rhd> t"
proof (cases C)
  case Hole with assms show ?thesis by simp
next
  case (More f ss1 D ss2)
  with assms have s: "s = Fun f (ss1 @ D\<langle>t\<rangle> # ss2)" by simp
  have "D\<langle>t\<rangle> \<in> set (ss1 @ D\<langle>t\<rangle> # ss2)" by simp
  hence "s \<rhd> D\<langle>t\<rangle>" unfolding s ..
  also have "D\<langle>t\<rangle> \<unrhd> t" by (induct D) auto
  finally show "s \<rhd> t" .
qed

lemma supt_not_refl[elim!]: assumes "t \<rhd> t" shows False
proof -
  from assms have "t \<noteq> t" by auto
  thus False by simp
qed

lemma supt_ctxt_conv: "(s \<rhd> t) = (\<exists>C. C \<noteq> \<box> \<and> s = C\<langle>t\<rangle>)" (is "_ = ?rhs")
proof
  assume "s \<rhd> t"
  hence "s \<unrhd> t" and "s \<noteq> t" by auto
  from `s \<unrhd> t` obtain C where "s = C\<langle>t\<rangle>" by auto
  with `s \<noteq> t` have "C \<noteq> \<box>" by auto
  with `s = C\<langle>t\<rangle>` show "?rhs" by auto
next
  assume "?rhs" thus "s \<rhd> t" by auto
qed

text {*
By adding \emph{non-empty} to statements \ref{B} and \ref{C} a similar characterisation for
proper subterms is obtained:
\begin{enumerate}
 \item @{text "s \<rhd> t"}
 \item @{text "\<exists>i p. i#p\<in>poss s \<and> s|_i<#p = t"}
 \item @{text "\<exists>C. C \<noteq> \<box> \<and> s = C\<langle>t\<rangle>"}
\end{enumerate}
*}

lemma supt_imp_subt_at_nepos: assumes "s \<rhd> t" shows "\<exists>i p. i<#p\<in>poss s \<and> s|_i<#p = t"
proof -
  from assms have "s \<unrhd> t" and "s \<noteq> t" unfolding supt_supteq_conv by auto
  then obtain p where supteq: "p \<in> poss s" "s|_p = t" using supteq_imp_subt_at by best
  have "p \<noteq> \<epsilon>"
  proof
    assume "p = \<epsilon>" hence "s = t" using `s|_p = t` by simp
    thus False using `s \<noteq> t` by simp
  qed
  then obtain i q where "p = i<#q" by (cases p) simp
  with supteq show ?thesis by auto
qed

lemma arg_neq:
  fixes f::"'f" and ss::"('f, 'v)term list"
  assumes "i < length ss" and "ss!i = Fun f ss" shows "False"
proof -
  from `i < length ss` have "(ss!i) \<in> set ss" by auto
  with assms show False by simp
qed

lemma supteq_not_supt[elim]: assumes "s \<unrhd> t" and "\<not> (s \<rhd> t)" shows "s = t"
using assms by (induct) auto

lemma supteq_not_supt_conv[simp]: "op \<unrhd> - op \<rhd> = Id" by auto

lemma subt_at_nepos_neq: assumes "i<#p \<in> poss s" shows "s|_(i<#p) \<noteq> s"
proof (cases s)
  fix x assume "s = Var x"
  hence "i<#p \<notin> poss s" by simp
  with assms show ?thesis by simp
next
  fix f ss assume "s = Fun f ss" show ?thesis
  proof
    assume "s|_i<#p = s"
    from assms have "i < length ss" unfolding `s = Fun f ss` by auto
    hence "ss!i \<in> set ss" by simp
    hence "Fun f ss \<rhd> ss!i" by (rule supt.arg)
    hence "Fun f ss \<noteq> ss!i" unfolding supt_supteq_conv by simp
    from `s|_i<#p = s` and assms
      have "ss!i \<unrhd> Fun f ss" using subt_at_imp_supteq' unfolding `s = Fun f ss` by auto
    with supt_not_sym[OF `Fun f ss \<rhd> ss!i`] have "ss!i = Fun f ss" by auto
    with `i < length ss` show False by (rule arg_neq)
  qed
qed

lemma subt_at_nepos_imp_supt: assumes "i<#p \<in> poss s" shows "s \<rhd> s|_i<#p"
proof -
  from assms have "s \<unrhd> s|_i<#p" by (rule subt_at_imp_supteq)
  from assms have "s|_i<#p \<noteq> s" by (rule subt_at_nepos_neq)
  from `s \<unrhd> s|_i<#p` and `s|_i<#p \<noteq> s` show ?thesis by (auto simp: supt_supteq_conv)
qed

lemma subt_at_nepos_imp_nectxt:
  assumes "i<#p \<in> poss s" and "s|_i<#p = t" shows "\<exists>C. C \<noteq> \<box> \<and> C\<langle>t\<rangle> = s"
proof -
  from assms obtain C where "C\<langle>s|_i<#p\<rangle> = s" using subt_at_imp_ctxt by best
  from `i<#p \<in> poss s`
    have "t \<noteq> s" unfolding `s|_i<#p = t`[symmetric] using subt_at_nepos_neq by best
  from assms and `C\<langle>s|_i<#p\<rangle> = s` have "C\<langle>t\<rangle> = s" by simp
  have "C \<noteq> \<box>"
  proof
    assume "C = \<box>"
    with `C\<langle>t\<rangle> = s` have "t = s" by simp
    with `t \<noteq> s` show False by simp
  qed
  with `C\<langle>t\<rangle> = s` show ?thesis by auto
qed

lemma nectxt_imp_supt_ctxt: "C \<noteq> \<box> \<Longrightarrow> C\<langle>t\<rangle> \<rhd> t"
proof (induct C)
  case Hole thus ?case by auto
next
  case (More f ls C rs) thus ?case
  proof (cases "C = \<box>")
    case True
    hence a: "(More f ls C rs)\<langle>t\<rangle> = Fun f (ls @ t#rs)" by auto
    hence "t \<in> set (ls @ t#rs)" by auto
    hence "Fun f (ls@t#rs) \<rhd> t" by (rule supt.arg)
    with a show ?thesis by auto
  next
    case False
    from `C \<noteq> \<box>` and `C \<noteq> \<box> \<Longrightarrow> C\<langle>t\<rangle> \<rhd> t` have b: "C\<langle>t\<rangle> \<rhd> t" by simp
    have c: "(More f ls C rs)\<langle>t\<rangle> = Fun f (ls @ C\<langle>t\<rangle>#rs)" by auto
    have d: "C\<langle>t\<rangle> \<in> set (ls @ C\<langle>t\<rangle>#rs)" by auto
    have "Fun f (ls @ C\<langle>t\<rangle>#rs) \<rhd> t"
      using supt.arg[where s = "C\<langle>t\<rangle>" and ss = "ls@C\<langle>t\<rangle>#rs",OF d] b by (rule supt_trans)
    with c show ?thesis by auto
  qed
qed

lemma supt_var: "\<not> (Var x \<rhd> u)"
proof
  assume "Var x \<rhd> u"
  then obtain C where "C \<noteq> \<box>" and "Var x = C\<langle>u\<rangle>" ..
  thus False by (cases C) auto
qed

lemma supt_const: "\<not> (Fun f [] \<rhd> u)"
proof
  assume "Fun f [] \<rhd> u"
  then obtain C where "C \<noteq> \<box>" and "Fun f [] = C\<langle>u\<rangle>" ..
  thus False by (cases C) auto
qed

lemma supteq_var_imp_eq:
  "(Var x \<unrhd> t) = (t = Var x)" (is "_ = (_ = ?x)")
proof
  assume "t = Var x" thus "Var x \<unrhd> t" by auto
next
  assume st: "?x \<unrhd> t"
  from st obtain C where "?x = C\<langle>t\<rangle>" by best
  thus "t = ?x" by (cases C) auto
qed

lemma Var_supt[elim!]: assumes "Var x \<rhd> t" shows P
using assms supt_var[of x t] by simp

lemma Fun_supt[elim]: assumes "Fun f ts \<rhd> s" obtains t where "t \<in> set ts" and "t \<unrhd> s"
using assms by (cases) (auto simp: supt_supteq_conv)

lemma subterm_induct:
  assumes "\<And>t. \<forall>s\<lhd>t. P s \<Longrightarrow> P t"
  shows [case_names subterm]: "P t"
proof -
  have "\<forall>s\<unlhd>t. P s"
  proof (induct t)
    case (Var x)
    have "\<forall>s\<lhd>Var x. P s" by blast
    with assms show ?case by blast
  next
    case (Fun f ts)
    have "\<forall>s\<lhd>Fun f ts. P s"
    proof (intro allI impI)
      fix s assume "Fun f ts \<rhd> s"
      then obtain t where "t \<in> set ts" and "t \<unrhd> s" ..
      with Fun show "P s" by simp
    qed
    with assms show ?case by blast
  qed
  thus ?thesis by auto
qed

text {*
A \emph{substitution} is a mapping~$\sigma$ from variables to terms.
We create a new data type in order to support instanciation of type classes.
*}
datatype ('f, 'v) subst = Subst "'v \<Rightarrow> ('f, 'v) term"

fun
  get_subst :: "('f, 'v) subst \<Rightarrow> ('v \<Rightarrow> ('f, 'v) term)"
where
  "get_subst (Subst s) = s"

fun
  subst_apply_term :: "('f, 'v) term \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) term" (infixl "\<cdot>" 60)
where
  "(Var x) \<cdot> \<sigma> = get_subst \<sigma> x" |
  "(Fun f ss) \<cdot> \<sigma> = Fun f (map (\<lambda>t. t \<cdot> \<sigma>) ss)"

fun
  subst_compose :: "('f, 'v) subst \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) subst" (infixl "\<circ>s" 65)
where
  "(Subst \<sigma>) \<circ>s \<tau> = Subst (\<lambda>x. (\<sigma> x) \<cdot> \<tau>)"

definition
  subst_empty :: "('f, 'v) subst" ("\<iota>")
where
  "\<iota> \<equiv> Subst (\<lambda>x. Var x)"

lemma subst_compose_apply_term_distrib[simp]: "t \<cdot> (\<sigma> \<circ>s \<tau>) = t \<cdot> \<sigma> \<cdot> \<tau>"
proof (induct t \<sigma> rule: subst_apply_term.induct)
  case (1 x \<sigma>) show ?case by (cases \<sigma>) simp_all
next
  case (2 f ss \<sigma>) thus ?case by auto
qed

lemma subst_apply_term_ident[simp]: "t \<cdot> \<iota> = t"
proof (induct t)
  case (Var x) show ?case by (simp add: subst_empty_def)
next
  case (Fun f ts)
  from map_ext[OF Fun[COMP impI]] show ?case by simp
qed

interpretation subst_monoid_mult: monoid_mult "\<iota>" "op \<circ>s"
proof
  fix \<sigma> \<tau> \<upsilon> show "\<sigma> \<circ>s \<tau> \<circ>s \<upsilon> = \<sigma> \<circ>s (\<tau> \<circ>s \<upsilon>)" by (cases \<sigma>) simp_all
next
  fix \<sigma> show "\<iota> \<circ>s \<sigma> = \<sigma>" by (cases \<sigma>) (simp_all add: subst_empty_def)
next
  fix \<sigma> show "\<sigma> \<circ>s \<iota> = \<sigma>" by (cases \<sigma>) simp_all
qed

declare subst_monoid_mult.mult_assoc[simp]

instantiation subst :: (type, type) monoid_mult begin
  definition [simp]: "1 \<equiv> \<iota>"
  definition [simp]: "op * \<equiv> op \<circ>s"
  instance by (intro_classes) simp_all
end

fun
  dom :: "('f, 'v) subst \<Rightarrow> 'v set"
where
  "dom (Subst \<sigma>) = {x. \<sigma>(x) \<noteq> (Var x)}"

fun
  ran :: "('f, 'v) subst \<Rightarrow> ('f, 'v) term set"
where
  "ran (Subst \<sigma>) = \<sigma> ` (dom (Subst \<sigma>))"

fun
  subst_restrict :: "('f, 'v) subst \<Rightarrow> 'v set \<Rightarrow> ('f, 'v) subst" ("_|'__" 60)
where
  "(Subst \<sigma>) |_ V = Subst (\<lambda>x. (if x \<in> V then \<sigma> x else (Var x)))"

lemma subst_empty_empty_dom[simp]: "dom \<iota> = {}" by (simp add: subst_empty_def)

lemma subst_empty_conv[simp]: "(dom \<sigma> = {}) = (\<sigma> = \<iota>)"
proof (rule iffI, cases \<sigma>)
 case (Subst s)
 assume "dom \<sigma> = {}"
 have "s = (\<lambda>x. Var x)"
 proof (rule ext)
  fix x show "s x = Var x"
  proof (rule ccontr)
   assume "s x \<noteq> Var x"
   hence "x \<in> dom \<sigma>" unfolding Subst by simp
   with `dom \<sigma> = {}` show False by simp
  qed
 qed
 thus "\<sigma> = \<iota>" unfolding Subst by (simp add: subst_empty_def)
next
 assume "\<sigma> = \<iota>" thus "dom \<sigma> = {}" by simp
qed

fun
  subst_apply_ctxt :: "('f, 'v) ctxt \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) ctxt" (infixl "\<cdot>c" 60)
where
  "\<box> \<cdot>c \<sigma> = \<box>" |
  "(More f ss1 D ss2) \<cdot>c \<sigma> = More f (map (\<lambda>t. t \<cdot> \<sigma>) ss1) (D \<cdot>c \<sigma>) (map (\<lambda>t. t \<cdot> \<sigma>) ss2)"

lemma subst_apply_term_ctxt_apply_distrib[simp]: "C\<langle>t\<rangle> \<cdot> \<mu> = (C \<cdot>c \<mu>)\<langle>t \<cdot> \<mu>\<rangle>"
by (induct C) auto

lemma subst_compose_ctxt_compose_distrib[simp]: "(C \<circ>c D) \<cdot>c \<sigma> = (C \<cdot>c \<sigma>) \<circ>c (D \<cdot>c \<sigma>)"
by (induct C) auto

lemma ctxt_compose_subst_compose_distrib[simp]: "C \<cdot>c (\<sigma> \<circ>s \<tau>) = (C \<cdot>c \<sigma>) \<cdot>c \<tau>"
by (induct C) auto

text {* Context substitutions. *}
datatype ('f, 'v) cs = CS "('f, 'v) ctxt" "('f, 'v) subst"

fun
  cs_apply :: "('f, 'v) term \<Rightarrow> ('f, 'v) cs \<Rightarrow> ('f, 'v) term" (infixl "\<cdot>cs" 60)
where
  "t \<cdot>cs (CS C \<mu>) = C\<langle>t \<cdot> \<mu>\<rangle>"

fun cs_compose :: "('f, 'v) cs \<Rightarrow> ('f, 'v) cs \<Rightarrow> ('f, 'v) cs" (infixl "\<circ>cs" 60)
where "CS C \<sigma> \<circ>cs CS D \<tau> = CS (D \<circ>c (C \<cdot>c \<tau>)) (\<sigma> \<circ>s \<tau>)"

lemma subst_apply_ctxt_ident[simp]: "C \<cdot>c \<iota> = C" by (induct C) simp_all

interpretation cs_monoid_mult: monoid_mult "CS \<box> \<iota>" "op \<circ>cs"
proof
  fix c d e show "c \<circ>cs d \<circ>cs e = c \<circ>cs (d \<circ>cs e)" by (cases c, cases d, cases e) simp_all
next
  fix c show "CS \<box> \<iota> \<circ>cs c = c" by (cases c) simp_all
next
  fix c show "c \<circ>cs CS \<box> \<iota> = c" by (cases c) simp_all
qed

declare cs_monoid_mult.mult_assoc[simp]

instantiation cs :: (type, type) monoid_mult begin
  definition [simp]: "1 \<equiv> CS \<box> \<iota>"
  definition [simp]: "op * \<equiv> op \<circ>cs"
  instance by (intro_classes) simp_all
end

lemma subst_power_compose_distrib: "\<mu> ^ (m + n) = (\<mu> ^ (m::nat) \<circ>s \<mu> ^ n)" by (induct m) simp_all

lemma cs_power_compose_distrib:
  "(CS C \<mu>) ^ (m + n) = (CS C \<mu>) ^ (m::nat) \<circ>cs (CS C \<mu>) ^ n"
by (induct m) simp_all

lemma cs_compose_apply_distrib[simp]: "t \<cdot>cs (c \<circ>cs d) = t \<cdot>cs c \<cdot>cs d"
by (cases c, cases d) simp_all

lemma term_cs_induct_power[simp]: "t \<cdot>cs (CS C \<mu>)^Suc(n::nat) = C\<langle>t \<cdot>cs (CS C \<mu>) ^ n \<cdot> \<mu>\<rangle>"
proof -
  have "t \<cdot>cs (CS C \<mu>)^(Suc n) = t \<cdot>cs (CS C \<mu>)^(n + 1)" by simp
  also have "\<dots> = t \<cdot>cs ((CS C \<mu>) ^n  \<circ>cs (CS C \<mu>))" unfolding cs_power_compose_distrib by simp
  also have "\<dots> = C\<langle>t \<cdot>cs (CS C \<mu>) ^ n \<cdot> \<mu>\<rangle>" by simp
  finally show ?thesis by simp
qed

text {* Properties of context-substitutions. *}
lemma cs_subst_distr[simp]: "t \<cdot>cs (CS C \<mu>) ^ n \<cdot> \<mu> = (t\<cdot>\<mu>) \<cdot>cs (CS (C \<cdot>c \<mu>) \<mu>) ^ n"
proof (induct n)
 case 0 thus ?case by (simp add: one_cs_def)
next
 case (Suc n)
 have "t \<cdot>cs (CS C \<mu>)^Suc n \<cdot> \<mu> = C\<langle>t \<cdot>cs (CS C \<mu>)^n \<cdot> \<mu>\<rangle> \<cdot> \<mu>" unfolding term_cs_induct_power by simp
 also have "\<dots> = C\<langle>(t\<cdot>\<mu>) \<cdot>cs (CS (C \<cdot>c \<mu>) \<mu>)^n\<rangle>\<cdot>\<mu>" unfolding Suc by simp
 also have "\<dots> = (C \<cdot>c \<mu>)\<langle>(t\<cdot>\<mu>) \<cdot>cs (CS (C \<cdot>c \<mu>) \<mu>)^n \<cdot> \<mu>\<rangle>" by simp
 finally show ?case unfolding term_cs_induct_power by simp
qed

lemma subt_at_PCons_distr[simp]:
 assumes "i<#p \<in> poss t" shows "t|_i<#p = (t|_<i>)|_p" using assms by (induct t) auto

lemma subt_at_append_distr[simp]: assumes "p<#>q \<in> poss t" shows "t|_(p<#>q) = (t|_p)|_q"
using assms proof (induct p arbitrary: t)
 case Empty thus ?case by simp
next
 case (PCons i p) thus ?case
 proof (induct t)
  case (Var x)
  have "poss(Var x) = {\<epsilon>}" by simp
  from Var have "i<#(p<#>q) = \<epsilon>" unfolding `poss(Var x) = {\<epsilon>}` by simp
  have "i<#(p<#>q) \<noteq> \<epsilon>" by simp
  from this `i<#(p<#>q) = \<epsilon>` show ?case by (rule notE)
 next
  case (Fun f ss) thus ?case by auto
 qed
qed

lemma empty_pos_in_poss[simp]: "\<epsilon> \<in> poss t" by (induct t) auto

lemma hole_pos_poss[simp]: "hole_pos C \<in> poss(C\<langle>t\<rangle>)" by (induct C) auto

lemma hole_pos_poss_conv: "((hole_pos C <#> q) \<in> poss(C\<langle>t\<rangle>)) = (q \<in> poss t)"
by (induct C) auto

lemma hole_pos_power_poss[simp]: "(hole_pos C)^(n::nat) \<in> poss((C^n)\<langle>t\<rangle>)"
by (induct n) (auto simp: hole_pos_poss_conv one_ctxt_def one_pos_def times_ctxt_def times_pos_def)

lemma subst_subt_at_distr[simp]: assumes "p \<in> poss t" shows "t|_p \<cdot> \<sigma> = (t \<cdot> \<sigma>)|_p"
using assms by (induct t arbitrary: p) auto

lemma subt_at_hole_pos[simp]: "C\<langle>t\<rangle>|_(hole_pos C) = t" by (induct C) auto

lemma poss_imp_subst_poss[simp]: "p \<in> poss t \<Longrightarrow> p \<in> poss(t\<cdot>\<sigma>)" by (induct t arbitrary: p) auto

lemma poss_imp_ctxt_subst_poss[simp]: assumes "p \<in> poss(C\<langle>t\<rangle>)" shows "p \<in> poss((C \<cdot>c \<sigma>)\<langle>t \<cdot> \<sigma>\<rangle>)"
proof -
 have "p \<in> poss(C\<langle>t\<rangle> \<cdot> \<sigma>)" by (rule poss_imp_subst_poss[OF assms])
 thus ?thesis by simp
qed

lemma cs_hole_ctxt_poss[simp]: "q \<in> poss t \<Longrightarrow> (hole_pos C <#> q) \<in> poss(t \<cdot>cs (CS C \<sigma>))"
by (simp add: hole_pos_poss_conv)

lemma cs_hole_poss_power_poss[simp]: "(hole_pos C)^(n::nat) \<in> poss(t \<cdot>cs (CS C \<sigma>)^n)" (is "?p^n \<in> poss(t \<cdot>cs ?c^n)")
proof (induct n)
 case 0 show ?case by (simp add: one_pos_def one_cs_def)
next
 case (Suc n) have "(?p <#> ?p^n) \<in> poss(t \<cdot>cs ?c^n \<cdot>cs ?c)" using cs_hole_ctxt_poss[OF Suc] by best
 hence "?p^(1+n) \<in> poss(t \<cdot>cs (?c^n \<circ>cs ?c^(1::nat)))" unfolding power_append_distr
   cs_compose_apply_distrib by (simp add: one_pos_def times_pos_def)
 thus ?case unfolding power_append_distr[symmetric] cs_power_compose_distrib[symmetric] by simp
qed

lemma subt_at_hole_pos_power[simp]:
 fixes t :: "('f,'v)term"
 shows "((t \<cdot>cs ((CS C \<mu>)^(n::nat)))|_((hole_pos C)^n)) = (t \<cdot> (\<mu>^n))" (is "?A n = ?B n")
proof (induct n)
 case 0 show ?case by (simp add: one_subst_def one_cs_def one_pos_def)
next
 case (Suc n)
 let ?p = "hole_pos C"
 let ?c = "CS C \<mu>"
 have "?p^Suc n \<in> poss(t \<cdot>cs ?c^Suc n)" by (rule cs_hole_poss_power_poss)
 hence "?p^(1 + n) \<in> poss(t \<cdot>cs ?c^(n + 1))" by simp
 hence A: "?p <#> ?p^n \<in> poss(C\<langle>t \<cdot>cs ?c^n \<cdot> \<mu>\<rangle>)"
   unfolding power_append_distr cs_power_compose_distrib
   by (simp add: times_pos_def times_cs_def one_pos_def one_cs_def)
 have "?A(Suc n) = ((t \<cdot>cs ((?c)^(n+1)))|_((?p)^(1+n)))" by simp
 also have "\<dots> = ((t \<cdot>cs ?c^n) \<cdot>cs ?c)|_(?p <#> (?p^n))"
   unfolding cs_power_compose_distrib power_append_distr
   by (simp add: times_cs_def one_cs_def one_pos_def times_pos_def)
 also have "\<dots> = C\<langle>t \<cdot>cs ?c^n \<cdot> \<mu>\<rangle>|_(?p <#> (?p^n))" by simp
 also have "\<dots> = (t \<cdot>cs ?c^n \<cdot> \<mu>)|_(?p^n)" unfolding subt_at_append_distr[OF A] by simp
 also have "\<dots> = (t \<cdot>cs ?c^n)|_(?p^n) \<cdot> \<mu>" by simp
 also have "\<dots> = (t \<cdot> \<mu>^n)\<cdot>\<mu>" unfolding Suc by simp
 also have "\<dots> = t \<cdot> (\<mu>^n \<circ>s \<mu>^1)" by (simp add: times_subst_def one_subst_def)
 finally show ?case unfolding subst_power_compose_distrib[symmetric] by simp
qed

definition
  subst :: "'v \<Rightarrow> ('f,'v) term \<Rightarrow> ('f,'v) subst" ("{_\<leftarrow>_}") 
where
  "{x \<leftarrow> t} = Subst (\<lambda>y. if x = y then t else Var y)"

lemma Var_subst[simp]: "Var x \<cdot> {x\<leftarrow>t} = t" by (simp add: subst_def)

lemma "Var x \<cdot> 1 \<cdot> 1 = Var x" by simp

inductive
  subst_instance :: "('f,'v)term \<Rightarrow> ('f,'v) term \<Rightarrow> bool" ("(_/ \<preceq> _)" [56, 56] 55)
where
  subst_instanceI[intro]: "s\<cdot>\<sigma> = t \<Longrightarrow> s \<preceq> t"

lemma subst_instance_trans[trans]:
 fixes s :: "('f,'v)term"
 assumes "s \<preceq> t" and "t \<preceq> u" shows "s \<preceq> u"
proof -
 from `s \<preceq> t` obtain \<sigma> where "s\<cdot>\<sigma> = t" by (cases rule: subst_instance.cases) best
 from `t \<preceq> u` obtain \<tau> where "t\<cdot>\<tau> = u" by (cases rule: subst_instance.cases) best
 hence "(s\<cdot>\<sigma>)\<cdot>\<tau> = u" unfolding `s\<cdot>\<sigma> = t` .
 hence "s\<cdot>(\<sigma> \<circ>s \<tau>) = u" by simp
 thus ?thesis by (rule subst_instanceI)
qed
 
lemma subst_instance_refl: "s \<preceq> s"
using subst_instanceI[where \<sigma> = "\<iota>" and s = s and t = s] by simp

lemma subst_neutral: "dom \<sigma> \<subseteq> V \<Longrightarrow> (Var x)\<cdot>(\<sigma>|_(V - {x})) = (Var x)" by (cases \<sigma>) auto

lemma subst_restrict_dom': "(\<lambda>x. if x \<in> dom (Subst s) then s x else (Var x)) x = s x"
proof (cases "x \<in> dom (Subst s)")
  case False thus ?thesis by simp
next
  case True thus ?thesis by simp
qed

lemma subst_restrict_dom[simp]: "\<sigma>|_dom \<sigma> = \<sigma>"
proof (cases \<sigma>)
 case (Subst s)
 have "\<sigma>|_dom \<sigma> = (Subst s)|_dom(Subst s)" unfolding Subst by simp
 also have "\<dots> = Subst(\<lambda>x. if x \<in> dom(Subst s) then s x else (Var x))" by simp
 also have "\<dots> = Subst s" unfolding subst_restrict_dom' by simp
 finally show ?thesis unfolding Subst by simp
qed

lemma subst_dom: assumes "x \<notin> dom (Subst s)" shows "s x = Var x"
using assms by auto

lemma subst_dom_neutral[simp]:
  fixes \<sigma>::"('f, 'v) subst"
  assumes "dom \<sigma> \<subseteq> V" shows "\<sigma>|_V = \<sigma>"
proof (cases \<sigma>)
 case (Subst s) {
  fix x
  have "(if x \<in> V then s x else Var x) = (if x \<in> dom(Subst s) then s x else Var x)"
  proof (cases "x \<in> dom \<sigma>")
    case True
    hence x: "x \<in> V = True" using assms by auto
    show ?thesis unfolding x if_True subst_restrict_dom' by (rule refl)
  next
    case False
    hence x: "x \<notin> dom (Subst s)" unfolding Subst .
    show ?thesis unfolding subst_dom[OF x] if_cancel by (rule refl)
  qed
 }
 hence "\<And>x.(if x \<in> V then s x else Var x) = (if x \<in> dom(Subst s) then s x else Var x)" .
 hence "\<And>x. (\<lambda>x. if x \<in> V then s x else Var x) x = (\<lambda>x. if x \<in> dom(Subst s) then s x else Var x) x" .
 hence "\<And>x. (\<lambda>x. if x \<in> V then s x else Var x) x = s x" unfolding subst_restrict_dom' .
 hence "(\<lambda>x. if x \<in> V then s x else Var x) = s" by (rule ext)
 hence "Subst (\<lambda>x. if x \<in> V then s x else Var x) = Subst s" by (simp only: subst.inject)
 hence "(Subst s)|_V = Subst s" by simp
 thus ?thesis unfolding Subst by simp
qed

lemma coincidence_lemma': "vars_term t \<subseteq> V \<Longrightarrow> t\<cdot>(\<sigma>|_vars_term t) = t\<cdot>(\<sigma>|_V)"
proof (induct t arbitrary: V)
 case (Var x) thus ?case by (induct \<sigma>) auto
next
 case (Fun f ss)
 let ?t = "Fun f ss"
 have "\<forall>s\<in>set ss. vars_term s \<subseteq> vars_term ?t" by auto
 hence A:"\<forall>s\<in>set ss. s\<cdot>(\<sigma>|_vars_term s) = s\<cdot>(\<sigma>|_vars_term ?t)" using Fun by simp
 hence "\<forall>s\<in>set ss. vars_term s \<subseteq> V" using Fun by auto
 hence "\<forall>s\<in>set ss. s\<cdot>(\<sigma>|_(vars_term s)) = s\<cdot>(\<sigma>|_V)" using Fun by simp
 with A have "\<forall>s\<in>set ss. s\<cdot>(\<sigma>|_vars_term ?t) = s\<cdot>(\<sigma>|_V)" by auto
 thus ?case by auto
qed

lemma [simp]: "\<sigma>|_UNIV = \<sigma>" by (cases \<sigma>) auto

lemma [simp]: "\<sigma>|_{} = \<iota>" by (cases \<sigma>) (simp_all add: subst_empty_def)

text {* The well-known coincidence lemma *}
lemma coincidence_lemma:"t\<cdot>\<sigma> = t\<cdot>(\<sigma>|_(vars_term t))"
proof -
 have "vars_term t \<subseteq> UNIV" by simp
 hence "t\<cdot>(\<sigma>|_vars_term t) = t\<cdot>\<sigma>" using coincidence_lemma'[where V = UNIV] by simp
 thus ?thesis by simp
qed

lemma dom_restrict:"dom(\<sigma>|_vars_term t) \<subseteq> vars_term t" by (cases \<sigma>) auto

lemma restrict_empty[simp]: assumes "x \<notin> dom \<sigma>" shows "\<sigma>|_{x} = \<iota>"
proof (cases \<sigma>)
 case (Subst s)
 have A: "\<And>x. x \<notin> dom \<sigma> \<Longrightarrow> s x = Var x" unfolding Subst by auto
 have "\<sigma>|_{x} = (Subst s)|_{x}" unfolding Subst by simp
 also have "\<dots> = Subst(\<lambda>y. if y \<in> {x} then s y else Var y)" by simp
 also have "\<dots> = Subst(\<lambda>y. if y = x then s y else Var y)" by simp
 also have "\<dots> = Subst(\<lambda>y. if y = x then s x else Var y)" by (simp cong: if_cong)
 also have "\<dots> = Subst(\<lambda>y. if y = x then Var x else Var y)" unfolding A[OF assms] by simp
 also have "\<dots> = Subst(\<lambda>y. if y = x then Var y else Var y)" by (simp cong: if_cong)
 also have "\<dots> = Subst(\<lambda>y. Var y)" by simp
 finally show ?thesis by (simp add: subst_empty_def)
qed

lemma restrict_empty': assumes "x \<notin> dom \<sigma>" and "\<sigma>|_V = \<iota>" shows "\<sigma>|_({x} \<union> V) = \<iota>"
proof (cases \<sigma>)
 case (Subst s)
 have "Subst (\<lambda>y. if y \<in> V then s y else Var y) = Subst (\<lambda>y. Var y)"
   using `\<sigma>|_V = \<iota>` unfolding Subst by (simp add: subst_empty_def)
 hence "(\<lambda>y. if y \<in> V then s y else Var y) = (\<lambda>y. Var y)" by simp
 hence A: "\<And>y. (if y \<in> V then s y else Var y) = Var y" by (rule fun_cong)
 {
  fix y
  have "(if y \<in> {x}\<union>V then s y else Var y) = Var y" 
  proof (cases "y = x")
   assume "y = x" thus ?thesis using `x \<notin> dom \<sigma>` unfolding Subst by auto
  next
   assume "y \<noteq> x" thus ?thesis using A by simp
  qed
 }
 hence "\<And>y. (if y \<in> {x}\<union>V then s y else Var y) = Var y" by simp
 thus ?thesis unfolding Subst by (auto simp: subst_empty_def)
qed

lemma restrict_empty_set:"finite V \<Longrightarrow> V \<inter> dom \<sigma> = {} \<Longrightarrow> \<sigma>|_V = \<iota>"
proof(induct rule: finite.induct)
 case emptyI thus ?case by simp
next
 case (insertI V x)
 hence "V \<inter> dom \<sigma> = {}" by simp
 with insertI have "\<sigma>|_V = \<iota>" by simp
 thus ?case using insertI restrict_empty'[where \<sigma> = \<sigma> and x = x and V = V] by simp
qed

lemma restrict_var:"x \<noteq> y \<Longrightarrow> (Var y)\<cdot>(\<sigma>|_(UNIV-{x})) = Var y\<cdot>\<sigma>" by (cases \<sigma>) auto


lemma vars_subst_ex: assumes A: "vars_term s \<subseteq> vars_term t" and B: "t\<cdot>\<sigma> = t\<cdot>\<tau>" shows "s\<cdot>\<sigma> = s\<cdot>\<tau>"
using A proof (induct s)
 case (Var x) thus ?case using B by (induct t) auto
next
 case (Fun f ss) thus ?case by auto
qed

lemma supteq_stable:
 fixes s t :: "('f,'v)term" and \<sigma> :: "('f, 'v) subst"
 assumes "s \<unrhd> t" shows "s\<cdot>\<sigma> \<unrhd> t\<cdot>\<sigma>"
using assms
proof (induct rule: supteq.induct)
 case (refl s) show ?case by (rule refl)
next
 case (subt u ss t f)
 from `u \<in> set ss` have "u\<cdot>\<sigma> \<in> set (map (\<lambda>t. t\<cdot>\<sigma>) ss)" "u\<cdot>\<sigma> \<unrhd> u\<cdot>\<sigma>" by auto
 hence "(Fun f ss)\<cdot>\<sigma> \<unrhd> u\<cdot>\<sigma>" unfolding subst_apply_term.simps by blast
 from this and `u\<cdot>\<sigma> \<unrhd> t\<cdot>\<sigma>` show ?case by (rule supteq_trans)
qed

lemma supt_stable: fixes \<sigma> :: "('f,'v)subst" assumes "s \<rhd> t" shows "s\<cdot>\<sigma> \<rhd> t\<cdot>\<sigma>"
using assms
proof (induct rule: supt.induct)
 case (arg s ss f) 
 hence "s\<cdot>\<sigma> \<in> set (map (\<lambda>t. t\<cdot>\<sigma>) ss)" by simp
thus ?case unfolding subst_apply_term.simps by (rule supt.arg)
next
 case (subt u ss t f)
 from `u \<in> set ss` have "u\<cdot>\<sigma> \<in> set (map (\<lambda>t. t\<cdot>\<sigma>) ss)" by simp
 hence "(Fun f ss)\<cdot>\<sigma> \<rhd> u\<cdot>\<sigma>" unfolding subst_apply_term.simps by (rule supt.arg)
 from this and `u\<cdot>\<sigma> \<rhd> t\<cdot>\<sigma>` show ?case by (rule supt_trans)
qed

text {* The set of all well-formed terms w.r.t. a given signature. Variables are
arbitrary *}
inductive_set wf_terms
  for F :: "('f\<times>nat)set"
where
  Var[intro]: "Var x \<in> wf_terms F" |
  Fun[intro]: "\<lbrakk>(f,length ts) \<in> F; \<And>t. t \<in> set ts \<Longrightarrow> t \<in> wf_terms F\<rbrakk> \<Longrightarrow> Fun f ts \<in> wf_terms F"

fun
  wf_terms_list :: "('f \<times> nat) list \<Rightarrow> ('f, 'v) term \<Rightarrow> bool"
where
  "wf_terms_list _ (Var _) = True" |
  "wf_terms_list F (Fun f ts) = ((f,length ts) \<in> set F \<and> list_all (wf_terms_list F) ts)"

lemma wf_terms_list_sound[simp]: "wf_terms_list F t = (t \<in> wf_terms (set F))"
proof (induct t)
  case (Var x) thus ?case by auto
next
  case (Fun f ts)
  note IH = this
  show ?case (is "?l = ?r")
  proof
    assume ?l
    with IH have "\<forall> t \<in> set ts. t \<in> wf_terms (set F)"
      and "(f,length ts) \<in> set F" by (auto simp: list_all_iff)
    thus ?r by auto
  next
    assume ?r    
    thus ?l 
    proof (cases rule: wf_terms.cases)
      case Fun
      with IH show ?thesis by (auto simp: list_all_iff)
    qed
  qed
qed

declare wf_terms_list.simps[simp del]

fun
  clean_term :: "'f sig \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term"
where
  "clean_term F (Var y) = Var y" |
  "clean_term F (Fun f ts) = (if (f, length ts) \<in> F
     then Fun f (map (clean_term F) ts)
      else Var (SOME z. True))"

fun
  clean_subst :: "'f sig \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v) subst"
where
  "clean_subst F (Subst s) = Subst (clean_term F \<circ> s)"

lemma clean_term_in_wf_terms[simp]: "clean_term F t \<in> wf_terms F"
proof (induct t)
  case (Var x) show ?case by auto
next
  case (Fun f ts)
  show ?case by (cases "(f,  length ts) \<in> F") (force simp: Fun, auto)
qed

lemma clean_term_idemp[simp]:
  "clean_term F (clean_term F t) = clean_term F t"
by (induct t) (auto)

lemma clean_subst_apply_term[simp]:
  assumes "t \<in> wf_terms F"
  shows "clean_term F (t \<cdot> \<sigma>) = t \<cdot> (clean_subst F \<sigma>)"
using assms proof (induct t rule: term_induct)
  case (Var x) show ?case by (cases \<sigma>) simp_all
next
  case (Fun f ss)
  from `Fun f ss \<in> wf_terms F`
    have "\<forall>t\<in>set ss. t \<in> wf_terms F" by (cases rule: wf_terms.cases) simp_all
  with Fun have IH: "\<And>t. t \<in> set ss \<Longrightarrow> clean_term F (t \<cdot> \<sigma>) = t \<cdot> (clean_subst F \<sigma>)" by best
  show ?case
  proof (cases "(f,length ss) \<in> F")
    case True show ?thesis by (auto simp: o_def IH True)
  next
    case False with `Fun f ss \<in> wf_terms F` have False by cases auto
    thus ?thesis by simp
  qed
qed

lemma wf_terms_subset[simp]: assumes "F \<subseteq> G" shows "wf_terms F \<subseteq> wf_terms G"
proof
  fix t assume "t \<in> wf_terms F" thus "t \<in> wf_terms G" using assms by (induct t) auto
qed

lemma wf_terms_funas_term[simp]: shows "t \<in> wf_terms (funas_term t)"
proof (induct t)
  case (Var x) show ?case by auto
next
  case (Fun f ss)
  have "\<forall>s\<in>set ss. s \<in> wf_terms (funas_term (Fun f ss))"
  proof
    fix s assume "s \<in> set ss"
    hence "funas_term s \<subseteq> funas_term (Fun f ss)" by auto
    moreover with Fun and `s \<in> set ss` have "s \<in> wf_terms (funas_term s)" by simp
    ultimately show "s \<in> wf_terms (funas_term (Fun f ss))" using wf_terms_subset by auto
  qed
  moreover have "(f,length ss) \<in> funas_term(Fun f ss)" by simp
  ultimately show ?case by fast
qed

fun
  funas_ctxt :: "('f, 'v) ctxt \<Rightarrow> 'f sig"
where
  "funas_ctxt Hole = {}" |
  "funas_ctxt (More f ss1 D ss2) = {(f, Suc (length (ss1 @ ss2)))}
     \<union> funas_ctxt D \<union> \<Union>(set (map funas_term (ss1 @ ss2)))"

lemma funas_term_ctxt_apply[simp]: "funas_term (C\<langle>t\<rangle>) = funas_ctxt C \<union> funas_term t"
proof (induct t)
  case (Var x) show ?case by (induct C) auto
next
  case (Fun f ts) show ?case by (induct C arbitrary: f ts) auto
qed

lemma funas_term_subst_apply:
  "funas_term (t\<cdot>\<sigma>) = funas_term t \<union> (\<Union>x\<in>vars_term t. funas_term (Var x \<cdot> \<sigma>))"
by (induct t) auto

lemma funas_term_wf_terms:
  assumes "funas_term t \<subseteq> F"
  shows "t \<in> wf_terms F"
using wf_terms_subset[OF assms] wf_terms_funas_term by auto

inductive_set
  wf_ctxts
  for F::"'f sig"
where
  Hole[intro]: "Hole \<in> wf_ctxts F" |
  More[intro]: "(f, Suc (length (ss1 @ ss2))) \<in> F \<Longrightarrow>
                (\<And>t. t \<in> set (ss1 @ ss2) \<Longrightarrow> t \<in> wf_terms F) \<Longrightarrow>
                D \<in> wf_ctxts F \<Longrightarrow> More f ss1 D ss2 \<in> wf_ctxts F"

lemma supteq_Var_id[simp]: assumes "Var x \<unrhd> s" shows "s = Var x"
using assms proof (induct s rule: term_induct)
  case (Var y) thus ?case by (cases rule: supteq.cases) simp_all
next
  case (Fun f ts)
  from `Var x \<unrhd> Fun f ts` have False by (cases rule: supteq.cases) 
  thus ?case by simp
qed

lemma supteq_imp_wf_terms: assumes "t \<in> wf_terms F" and "t \<unrhd> s" shows "s \<in> wf_terms F"
using assms proof (induct t)
  case (Var x) from supteq_Var_id[OF this] show ?case by auto
next
  case (Fun f ts)
  from `Fun f ts \<unrhd> s`
    have "s = Fun f ts \<or> (\<exists>u\<in>set ts. u \<unrhd> s)" by (cases rule: supteq.cases) auto
  thus ?case proof
    assume "s = Fun f ts" thus ?thesis using Fun by auto
  next
    assume "\<exists>u\<in>set ts. u \<unrhd> s"
    then obtain u where "u \<in> set ts" and "u \<unrhd> s" by auto
    with Fun show ?thesis by auto
  qed
qed

lemma arg_not_term[simp]: assumes "t \<in> set ts" shows "Fun f ts \<noteq> t"
proof (rule ccontr)
  assume "\<not> Fun f ts \<noteq>  t"
  hence "size(Fun f ts) = size t" by simp
  moreover from assms have "size t < list_size size ts" by (induct ts) auto
  ultimately show False by simp
qed

lemma arg_subteq[simp]: assumes "t \<in> set ts" shows "Fun f ts \<unrhd> t"
using assms by auto
  
fun
  args :: "('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "args (Var x) = []" |
  "args (Fun f ts) = ts"

fun
  num_args :: "('f, 'v) term \<Rightarrow> nat"
where
  "num_args (Var x) = 0" |
  "num_args (Fun f ts) = length ts"

lemma supt_imp_args:
  assumes "\<forall>t. s \<rhd> t \<longrightarrow> P t" shows "\<forall>t\<in>set(args s). P t"
using assms by (cases s) (simp_all add: supt_supteq_conv)

lemma ctxt_apply_eq_False[simp]:
  assumes "t = (More f ss1 D ss2)\<langle>t\<rangle>" (is "_ = ?C\<langle>_\<rangle>") shows "False"
proof -
  have "?C \<noteq> \<box>" by auto
  from this and  assms have "t \<rhd> t" by (rule ctxt_supt)
  thus False by auto
qed

lemma wf_terms_Fun[simp]: assumes "Fun f ts \<in> wf_terms F" shows "(f,length ts) \<in> F"
using assms by cases auto

lemma wf_terms_clean_term_ident[simp]:
  assumes "t \<in> wf_terms F" shows "clean_term F t = t"
using assms proof (induct t rule: term_induct)
  case (Var x) show ?case by simp
next
  case (Fun f ts)
  hence IH: "\<And>s. s \<in> set ts \<Longrightarrow> clean_term F s = s" by (auto simp: supteq_imp_wf_terms)
  thus ?case
  proof (cases "(f, length ts) \<in> F")
    case False with Fun show ?thesis by simp
  next
    case True thus ?thesis using map_idI[OF IH, of ts] by auto
  qed
qed

lemma supteq_imp_funas_subset[simp]:
  "s \<unrhd> t \<Longrightarrow> funas_term t \<subseteq> funas_term s"
by (induct rule: supteq.induct) auto

lemma subst_apply_ctxt_clean_subst_wf_ctxts:
  assumes "C \<in> wf_ctxts F"
  shows "C \<cdot>c (clean_subst F \<sigma>) \<in> wf_ctxts F"
using assms proof (induct C rule: ctxt.induct)
  case Hole show ?case by auto
next
  case (More f ss1 E ss2)
  let ?C = "More f ss1 E ss2"
  from `?C \<in> wf_ctxts F`
    have "(f, Suc (length (ss1 @ ss2))) \<in> F \<and> (\<forall>t\<in>set (ss1 @ ss2). t \<in> wf_terms F)"
    by cases auto
  hence in_F: "(f, Suc (length (ss1 @ ss2))) \<in> F"
    and all: "\<forall>t\<in>set (ss1 @ ss2). t \<in> wf_terms F" by auto
  from in_F
    have in_F': "(f, Suc (length ((map (\<lambda>t. t \<cdot> (clean_subst F \<sigma>)) ss1)
      @ (map (\<lambda>t. t \<cdot> (clean_subst F \<sigma>)) ss2)))) \<in> F" by simp
  from all have all': "\<forall>t\<in>set(ss1 @ ss2). t \<cdot> (clean_subst F \<sigma>) = clean_term F (t \<cdot> \<sigma>)" by auto
  have "\<forall>t\<in>set (ss1 @ ss2). clean_term F (t \<cdot> \<sigma>) \<in> wf_terms F" by auto
  with all' have "\<forall>t\<in>set (ss1 @ ss2). t \<cdot> (clean_subst F \<sigma>) \<in> wf_terms F" by auto
  hence "\<forall>t\<in>set (map (\<lambda>t. t \<cdot> (clean_subst F \<sigma>)) (ss1 @ ss2)). t \<in> wf_terms F" by auto
  hence args: "\<And>t. t \<in> set (map (\<lambda>t. t \<cdot> (clean_subst F \<sigma>)) ss1
    @ map (\<lambda>t. t \<cdot> (clean_subst F \<sigma>)) ss2) \<Longrightarrow> t \<in> wf_terms F" by auto
  from `?C \<in> wf_ctxts F` have "E \<in> wf_ctxts F" by cases auto
  with More have IH: "E \<cdot>c (clean_subst F \<sigma>) \<in> wf_ctxts F" by best
  from args wf_ctxts.More[OF in_F' _ this]
    show ?case unfolding subst_apply_ctxt.simps by best
qed

lemma ctxt_compose_wf[simp]:
  "C \<in> wf_ctxts F \<Longrightarrow> E \<in> wf_ctxts F \<Longrightarrow> (C \<circ>c E) \<in> wf_ctxts F"
proof (induct C)
  case Hole thus ?case by simp
next
  case (More f ss1 G ss2)
  let ?C = "More f ss1 G ss2"
  from `?C \<in> wf_ctxts F` have G: "G \<in> wf_ctxts F" by cases auto
  from `?C \<in> wf_ctxts F`
    have "(f, Suc (length (ss1 @ ss2))) \<in> F \<and> (\<forall>t\<in>set (ss1 @ ss2). t \<in> wf_terms F)"
    by cases auto
  hence in_F: "(f, Suc (length (ss1 @ ss2))) \<in> F"
    and args: "\<And>t. t \<in> set (ss1 @ ss2) \<Longrightarrow> t \<in> wf_terms F" by auto
  from G More have IH: "(G \<circ>c E) \<in> wf_ctxts F" by simp
  from wf_ctxts.More[OF in_F _ IH] args show ?case by auto
qed

lemma set_set_cons:
  assumes "P x" and set:"\<And>y. y \<in> set xs \<Longrightarrow> P y" shows "y \<in> set(x#xs) \<Longrightarrow> P y"
using assms by auto

lemma ctxt_apply_wf[simp]: 
  shows "C \<in> wf_ctxts F \<Longrightarrow> t \<in> wf_terms F \<Longrightarrow> (C\<langle>t\<rangle>) \<in> wf_terms F"
proof (induct C)
  case Hole thus ?case by simp
next
  case (More f ss1 E ss2)
  let ?C = "More f ss1 E ss2"
  from `?C \<in> wf_ctxts F` have E: "E \<in> wf_ctxts F" by cases auto
  with More have IH: "E\<langle>t\<rangle> \<in> wf_terms F" by simp
  from `?C \<in> wf_ctxts F`
    have "(f, Suc (length (ss1 @ ss2))) \<in> F \<and> (\<forall>t\<in>set (ss1 @ ss2). t \<in> wf_terms F)"
    by cases auto
  hence in_F: "(f, Suc (length (ss1 @ ss2))) \<in> F"
    and args: "\<And>t. t \<in> set (ss1 @ ss2) \<Longrightarrow> t \<in> wf_terms F" by auto
  have len: "length(ss1@E\<langle>t\<rangle>#ss2) = Suc(length(ss1@ss2))" by auto
  from in_F have in_F': "(f,length(ss1@E\<langle>t\<rangle>#ss2)) \<in> F" unfolding len .
  from args have pre: "\<forall>t\<in>set ss1. t \<in> wf_terms F" by simp
  from args have "\<And>t. t \<in> set ss2 \<Longrightarrow> t \<in> wf_terms F" by simp
  with set_set_cons[of "\<lambda>t. t \<in> wf_terms F" "E\<langle>t\<rangle>" ss2, OF IH this]
    have "\<forall>u\<in>set (E\<langle>t\<rangle> # ss2). u \<in> wf_terms F" by blast
  with pre have a: "\<forall>u\<in>set (ss1 @ E\<langle>t\<rangle> # ss2). u \<in> wf_terms F" by auto
  have "\<And>u. u \<in> set (ss1 @ E\<langle>t\<rangle> # ss2) \<Longrightarrow> u \<in> wf_terms F" using a by best
  with wf_terms.Fun[OF in_F'] show ?case by auto
qed

lemma subst_apply_term_clean_wf_terms[simp]:
  assumes "t \<in> wf_terms F" shows "t \<cdot> (clean_subst F \<sigma>) \<in> wf_terms F"
proof -
  from assms have clean: "clean_term F (t \<cdot> \<sigma>) = t \<cdot> (clean_subst F \<sigma>)" by auto
  have "clean_term F (t \<cdot> \<sigma>) \<in> wf_terms F" by simp
  thus ?thesis unfolding clean .
qed

lemma ctxt_apply_wf_imp_ctxt_and_term_wf:
  assumes "C\<langle>t\<rangle> \<in> wf_terms F" shows "C \<in> wf_ctxts F \<and> t \<in> wf_terms F"
using assms proof (induct C)
  case Hole thus ?case by auto
next
  case (More f ss1 E ss2)
  let ?C = "More f ss1 E ss2"
  from `?C\<langle>t\<rangle> \<in> wf_terms F`
    have "(f, length (ss1 @ E\<langle>t\<rangle> # ss2)) \<in> F \<and> (\<forall>s\<in>set (ss1 @ E\<langle>t\<rangle> # ss2). s \<in> wf_terms F)"
    unfolding ctxt_apply.simps by cases auto
  hence in_F: "(f,length(ss1@E\<langle>t\<rangle>#ss2)) \<in> F"
    and args: "\<forall>s\<in>set(ss1@E\<langle>t\<rangle>#ss2). s \<in> wf_terms F" by auto
  from in_F have in_F': "(f,Suc(length(ss1@ss2))) \<in> F" by auto
  from args have "E\<langle>t\<rangle> \<in> wf_terms F" by simp
  with More have "E \<in> wf_ctxts F \<and> t \<in> wf_terms F" by best
  hence "E \<in> wf_ctxts F" and "t \<in> wf_terms F" by auto
  from args have "\<forall>s\<in>set(ss1). s \<in> wf_terms F" by auto
  moreover from args have "\<forall>s\<in>set(ss2). s \<in> wf_terms F" by auto
  ultimately have "\<forall>s\<in>set(ss1@ss2). s \<in> wf_terms F" by auto
  hence "\<And>s. s \<in> set(ss1@ss2) \<Longrightarrow> s \<in> wf_terms F" by auto
  from wf_ctxts.More[OF in_F' _ `E \<in> wf_ctxts F`] this `t \<in> wf_terms F` show ?case by auto
qed

lemma clean_ctxt_apply:
  assumes "C \<in> wf_ctxts F" shows "clean_term F (C\<langle>t\<rangle>) = C\<langle>clean_term F t\<rangle>"
using assms proof (induct C)
  case Hole thus ?case by simp
next
  case (More f ss1 ss2 E)
  let ?C = "More f ss1 E ss2"
  from More have "(f,Suc(length(ss1@ss2))) \<in> F" by auto
  hence in_F: "(f, length (ss1 @ E\<langle>clean F t\<rangle> # ss2)) \<in> F" by auto
  from More have "clean_term F (?C\<langle>t\<rangle>) = Fun f (map (clean_term F) (ss1 @ E\<langle>t\<rangle> # ss2))" by auto
  from `\<And>t. t \<in> set(ss1 @ ss2) \<Longrightarrow> t \<in> wf_terms F`
    have args: "\<forall>t\<in>set (ss1 @ ss2). clean_term F t = t" by auto
  hence "\<forall>t\<in>set ss1. clean_term F t = t" by auto
  hence pre: "map (clean_term F) ss1 = ss1" using map_idI[of ss1 "clean_term F"] by best
  from args have "\<forall>t\<in>set ss2. clean_term F t = t" by auto
  hence "map (clean_term F) ss2 = ss2" using map_idI[of ss2 "clean_term F"] by best
  with `clean_term F (E\<langle>t\<rangle>) = E\<langle>clean_term F t\<rangle>`
    have post: "map (clean_term F) (E\<langle>t\<rangle> # ss2) = E\<langle>clean_term F t\<rangle> # ss2" by auto
  have "map (clean_term F) (ss1 @ E\<langle>t\<rangle> # ss2)
    = map (clean_term F) ss1 @ map (clean_term F) (E\<langle>t\<rangle> # ss2)"
    by simp
  also have "\<dots> = ss1 @ E\<langle>clean_term F t\<rangle> # ss2" unfolding pre post by simp
  finally have "map (clean_term F) (ss1@ E\<langle>t\<rangle> # ss2) = ss1 @ E\<langle>clean_term F t\<rangle> # ss2" .
  with in_F show ?case by simp
qed

lemma ctxt_power_compose_distr:
  "C ^ (m + n) = C ^ m \<circ>c C ^ n"
by (induct m) (simp_all add: one_ctxt_def times_ctxt_def)

fun
  vars_ctxt :: "('f, 'v) ctxt \<Rightarrow> 'v set"
where
  "vars_ctxt \<box> = {}" |
  "vars_ctxt (More f ss1 D ss2) = vars_ctxt D \<union> \<Union>set (map vars_term (ss1 @ ss2))"

lemma subst_apply_id': assumes "vars_term t \<inter> V = {}" shows "t\<cdot>(\<sigma>|_V) = t"
using assms proof (induct t)
  case (Var x) thus ?case by (cases \<sigma>) simp_all
next
  case (Fun f ts)
  hence IH: "\<forall>s\<in>set ts. s\<cdot>(\<sigma>|_V) = s" by auto
  with map_idI[of ts "\<lambda>t. t\<cdot>(\<sigma>|_V)"] show ?case by simp
qed

lemma subst_apply_ctxt_id: assumes "vars_ctxt C \<inter> V = {}" shows "C \<cdot>c (\<sigma>|_V) = C"
using assms proof (induct C)
  case Hole show ?case by simp
next
  case (More f ss1 D ss2)
  hence IH: "D \<cdot>c (\<sigma>|_V) = D" by auto
  from More have "\<forall>s\<in>set(ss1@ss2). vars_term s \<inter> V = {}" by auto
  with subst_apply_id' have args: "\<forall>s\<in>set(ss1@ss2). s\<cdot>(\<sigma>|_V) = s" by best
  from args have "\<forall>s\<in>set ss1. s\<cdot>(\<sigma>|_V) = s" by simp
  with map_idI[of ss1 "\<lambda>t. t\<cdot>(\<sigma>|_V)"] have ss1: "map (\<lambda>s. s\<cdot>(\<sigma>|_V)) ss1 = ss1" by best
  from args have "\<forall>s\<in>set ss2. s\<cdot>(\<sigma>|_V) = s" by simp
  with map_idI[of ss2 "\<lambda>t. t\<cdot>(\<sigma>|_V)"] have ss2: "map (\<lambda>s. s\<cdot>(\<sigma>|_V)) ss2 = ss2" by best
  show ?case by (simp add: ss1 ss2 IH)
qed

lemma vars_term_Var_id: "vars_term o Var = (\<lambda>x. {x})" by (rule ext) simp

lemma ctxt_exhaust_rev[case_names Hole More]:
  assumes "C = \<box> \<Longrightarrow> P" and "\<And>D f ss1 ss2. C = D \<circ>c (More f ss1 \<box> ss2) \<Longrightarrow> P" shows "P"
proof (cases C)
  case Hole with assms show ?thesis by simp
next
  case (More g ts1 E ts2)
  hence "\<exists>D f ss1 ss2. C = D \<circ>c (More f ss1 \<box> ss2)"
  proof (induct E arbitrary: C g ts1 ts2)
    case Hole hence "C = \<box> \<circ>c (More g ts1 \<box> ts2)" by simp
    thus ?case by best
  next
    case (More h us1 F us2)
    from More(1)[of "More h us1 F us2"]
     obtain G i vs1 vs2 where IH: "More h us1 F us2 = G \<circ>c More i vs1 \<box> vs2" by force
    from More have "C = (More g ts1 \<box> ts2 \<circ>c G) \<circ>c More i vs1 \<box> vs2" unfolding IH by simp
    thus ?case by best
  qed
  thus ?thesis using assms by auto
qed

lemma ctxt_compose_wf_imp_ctxts_wf:
  assumes "C \<circ>c D \<in> wf_ctxts F" shows "C \<in> wf_ctxts F \<and> D \<in> wf_ctxts F"
using assms proof (induct C)
  case Hole thus ?case by auto
next
  case (More f ss1 E ss2)
  from More(2)
    have wf: "(f,Suc(length(ss1@ss2))) \<in> F \<and> (\<forall>t\<in>set(ss1@ss2). t \<in> wf_terms F)
      \<and> E \<circ>c D \<in> wf_ctxts F"
    unfolding ctxt_compose.simps by (cases rule:wf_ctxts.cases) auto
  with More have IH: "E \<in> wf_ctxts F \<and> D \<in> wf_ctxts F" by fast
  with wf show ?case by auto
qed

fun
  subst_extend :: "('f, 'v) subst \<Rightarrow> ('v \<times> ('f, 'v) term) list \<Rightarrow> ('f, 'v) subst"
where
  "subst_extend (Subst s) vts = Subst (\<lambda>x.
    (case lookup x vts of
      Some t \<Rightarrow> t
    | None   \<Rightarrow> s(x)))"

lemma lookup_None[simp]:
  assumes "V \<inter> set vs = {}" and "x \<in> V" shows "lookup x (zip vs ts) = None"
using assms proof (induct vs arbitrary: ts)
  case Nil show ?case by simp
next
  case (Cons v vs)
  note IH = this
  hence neq: "v \<noteq> x" by auto
  show ?case
  proof (cases ts)
    case Nil show ?thesis unfolding Nil by simp
  next
     case (Cons t ts')
     from IH have IH': "lookup x (zip vs ts') = None" by auto
     have "lookup x (zip (v#vs) ts) = lookup x ((v,t)#zip vs ts')" unfolding Cons by simp
     also have "\<dots> = lookup x (zip vs ts')" using neq by simp
     also have "\<dots> = None" unfolding IH' by simp
     finally show ?thesis .
  qed
qed

lemma subst_extend_id:
  assumes "V \<inter> set vs = {}" and "vars_term t \<subseteq> V"
  shows "t \<cdot> subst_extend \<sigma> (zip vs ts) = t \<cdot> \<sigma>"
using assms by (induct t,cases \<sigma>) auto

lemma lookup_Some[simp]:
  assumes "distinct vs" and "length vs \<le> length ts" and "i < length vs"
  shows "lookup (vs!i) (zip vs ts) = Some(ts!i)"
using assms proof (induct i arbitrary: vs ts)
  case 0 thus ?case by (induct ts) (induct vs,auto)+
next
  case (Suc i) show ?case
  proof (cases ts)
    case Nil with Suc show ?thesis unfolding Nil by simp
  next
    case (Cons t ts')
    note Cons' = this
    show ?thesis
    proof (cases vs)
      case Nil from Suc show ?thesis unfolding Cons' Nil by simp
    next
      case (Cons v vs') from Suc show ?thesis unfolding Cons Cons' by auto
    qed
  qed
qed

lemma subst_extend_absorb:
  assumes "distinct vs" and "length vs = length ss"
  shows "map (\<lambda>t. t\<cdot>(subst_extend \<sigma> (zip vs ss))) (map Var vs) = ss" (is "?ss = _")
proof -
  obtain s where \<sigma>: "\<sigma> = Subst s" by (cases \<sigma>) auto
  let ?\<sigma> = "subst_extend (Subst s) (zip vs ss)"
  from assms have "length vs \<le> length ss" by simp
  from assms have "length ?ss = length ss" by simp
  moreover have "\<forall>i<length ?ss. ?ss!i = ss!i"
  proof (intro allI impI)
    fix i assume "i < length ?ss"
    hence i: "i < length(map Var vs)" by simp
    hence len: "i < length vs" by simp
    have "?ss!i = (map Var vs!i)\<cdot>?\<sigma>" unfolding nth_map[OF i,of "\<lambda>t. t\<cdot>?\<sigma>"] \<sigma> by simp
    also have "\<dots> = Var(vs!i)\<cdot>?\<sigma>" unfolding nth_map[OF len] by simp
    also have "\<dots> = (case lookup (vs!i) (zip vs ss) of None \<Rightarrow> s(vs!i) | Some t \<Rightarrow> t)" by simp
    also have "\<dots> = ss!i" unfolding lookup_Some[OF `distinct vs` `length vs \<le> length ss` len] by simp
    finally show "?ss!i = ss!i" by simp
  qed
  ultimately show ?thesis by (rule nth_equalityI)
qed

lemma vars_term_finite[simp]: "finite(vars_term t)" by (induct t) simp_all

fun
  linear_term :: "('f, 'v) term \<Rightarrow> bool"
where
  [code del]: "linear_term (Var _) = True" |
  "linear_term (Fun _ ts) = (is_partition (map vars_term ts) \<and> list_all linear_term ts)"

fun map_term :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)term \<Rightarrow> ('g,'v)term"
where "map_term fg (Var x) = (Var x)"
    | "map_term fg (Fun f ts) = Fun (fg f) (map (map_term fg) ts)"

fun map_subst :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)subst \<Rightarrow> ('g,'v)subst"
where "map_subst fg (Subst sigma) = Subst (\<lambda> x. map_term fg (sigma x))"

fun map_ctxt :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('g,'v)ctxt"
where "map_ctxt fg \<box> = \<box>"
   |  "map_ctxt fg (More f ss C ts) = More (fg f) (map (map_term fg) ss) (map_ctxt fg C) (map (map_term fg) ts)"


lemma map_term_comp: "map_term fg (map_term gh t) = map_term (fg o gh) t"
  by (induct t, auto)

lemma map_subst_distrib[simp]: "map_term fg (t \<cdot> \<sigma>) = map_term fg t \<cdot> map_subst fg \<sigma>"
proof (cases \<sigma>)
  case (Subst sigma)
  have "map_term fg (t \<cdot> Subst sigma) = map_term fg t \<cdot> map_subst fg (Subst sigma)"    
    by (induct t, auto)
  with Subst show ?thesis by simp
qed

lemma map_term_id[simp]: "map_term id = id"
proof
  fix t
  show "map_term id t = id t" by (simp, induct t, auto simp: map_idI)
qed

lemma map_term_power: fixes f :: "'f \<Rightarrow> 'f" 
  shows "((map_term f) ^^ n) = map_term (f ^^ n)"
proof (rule sym, intro ext)
  fix t
  show "map_term (f ^^ n) t = (map_term f ^^ n) t"
  proof (induct n)
    case 0
    show ?case 
      by (simp, unfold id_def[symmetric] map_term_id, simp)
  next
    case (Suc n)
    show ?case by (simp add: Suc[symmetric] map_term_comp o_def)
  qed
qed

lemma map_ctxt_distrib[simp]: "map_term fg (C\<langle>t\<rangle>) = (map_ctxt fg C)\<langle>map_term fg t\<rangle>" 
  by  (induct C, auto)

fun map_term_wa :: "('f \<Rightarrow> nat \<Rightarrow> 'g) \<Rightarrow> ('f,'v)term \<Rightarrow> ('g,'v)term"
  where "map_term_wa fg (Var x) = (Var x)"
     |  "map_term_wa fg (Fun f ts) = Fun (fg f (length ts)) (map (map_term_wa fg) ts)"

fun map_ctxt_wa :: "('f \<Rightarrow> nat \<Rightarrow> 'g) \<Rightarrow> ('f,'v)ctxt \<Rightarrow> ('g,'v)ctxt"
  where "map_ctxt_wa fg \<box> = \<box>"
     |  "map_ctxt_wa fg (More f bef C aft) = More (fg f (Suc (length bef + length aft))) (map (map_term_wa fg) bef) (map_ctxt_wa fg C) (map (map_term_wa fg) aft)"

abbreviation map_subst_wa :: "('f \<Rightarrow> nat \<Rightarrow> 'g) \<Rightarrow> ('f,'v)subst \<Rightarrow> ('g,'v)subst"
  where "map_subst_wa fg \<sigma> \<equiv> Subst (\<lambda> x. map_term_wa fg (get_subst \<sigma> x))"

lemma map_term_wa_subst[simp]: "map_term_wa fg (t \<cdot> \<sigma>) = map_term_wa fg t \<cdot> map_subst_wa fg \<sigma>"
  by (induct t, auto)

lemma map_term_wa_ctxt[simp]: "map_term_wa fg (C\<langle>t\<rangle>) = (map_ctxt_wa fg C) \<langle>map_term_wa fg t\<rangle>"
  by (induct C, auto)


lemma map_term_wa_id: assumes id: "\<And> f n. fg f n = f"
  shows "map_term_wa fg t = t"
proof (induct t, simp)
  case (Fun f ss)
  thus ?case by (simp add: id, induct ss, auto)
qed

lemma clean_term_if[simp]:
  "clean_term F (if P then s else t) = (if P then clean_term F s else clean_term F t)"
by simp

declare size_simps[simp]
fun weak_match :: "('f,'v)term \<Rightarrow> ('f,'v)term \<Rightarrow> bool"
where "weak_match _ (Var _) = True"
  |   "weak_match (Var _) (Fun _ _) = False"
  |   "weak_match (Fun f ts) (Fun g ss) = (f = g \<and> length ts = length ss \<and> (\<forall> i < length ts. weak_match (ts ! i) (ss ! i)))"
declare size_simps[simp del]

lemma weak_match_refl: "weak_match t t"
  by (induct t, auto)

lemma weak_match_match: "weak_match (t \<cdot> \<sigma>) t"
  by (induct t, auto)

lemma clean_weak_match: "weak_match t (clean_term F t)"
proof (induct t, simp)
  case (Fun f ts)
  let ?n = "length ts"
  show ?case
  proof (cases "(f,?n) \<in> F")
    case False
    thus ?thesis by simp
  next
    case True
    {
      fix i
      assume i: "i < ?n"
      hence "ts ! i \<in> set ts" by simp
      from Fun[OF this] have "weak_match (ts ! i) (clean_term F (ts ! i))" .
    }
    with True show ?thesis by simp
  qed
qed

lemma linear_weak_match:
assumes "linear_term l" and "weak_match t s" and "s = l \<cdot> \<sigma>"
shows "\<exists>\<tau>. t = l\<cdot>\<tau> \<and> (\<forall> x \<in> vars_term l. weak_match (Var x \<cdot> \<tau>) (Var x \<cdot> \<sigma>))"
using assms proof (induct l arbitrary: s t)
  case (Var x)
  show ?case
  proof (rule exI[of _ "Subst (\<lambda> y. t)"], intro conjI, simp)
    from Var show "\<forall>x\<in>vars_term (Var x). weak_match (Var x \<cdot> Subst (\<lambda>y. t)) (Var x \<cdot> \<sigma>)"
      by force
  qed
next
  case (Fun f ls)
  let ?n = "length ls"
  from Fun(4) obtain ss where s: "s = Fun f ss" and lss: "length ss = ?n"  by (cases s, auto)
  with Fun(4) have match: "\<And> i. i < ?n \<Longrightarrow> ss ! i = (ls ! i) \<cdot> \<sigma>" by auto
  from Fun(3) s lss obtain ts where t: "t = Fun f ts" and lts: "length ts = ?n" by (cases t, auto)
  with Fun(3) s have weak_match: "\<And> i. i < ?n \<Longrightarrow> weak_match (ts ! i) (ss ! i)" by auto
  from Fun(2) have linear: "\<And> i. i < ?n \<Longrightarrow> linear_term (ls ! i)" unfolding linear_term.simps list_all_iff by auto
  let ?cond = "\<lambda> \<tau> i. ts ! i = ls ! i \<cdot> \<tau> \<and> (\<forall>x\<in>vars_term (ls ! i). weak_match (Var x \<cdot> \<tau>) (Var x \<cdot> \<sigma>))"
  {
    fix i
    assume i: "i < ?n"
    hence "ls ! i \<in> set ls" by simp
    from Fun(1)[OF this linear[OF i] weak_match[OF i] match[OF i]]
    have "\<exists> \<tau>. ?cond \<tau> i" .
  }
  hence "\<forall> i. \<exists> \<tau>. (i < ?n \<longrightarrow> ?cond \<tau> i)" by auto
  from choice[OF this] obtain subs where subs: "\<And> i. i < ?n \<Longrightarrow> ?cond (subs i) i" by auto  
  obtain \<tau> where \<tau>: "\<tau> = Subst (\<lambda> x. Var x \<cdot> subs (THE i. i < ?n \<and> x \<in> vars_term (ls ! i)))"
    by auto
  from Fun(2) have distinct: "is_partition(map vars_term ls)" by simp
  {
    fix i x
    assume i: "i < ?n" and x: "x \<in> vars_term (ls ! i)"
    have "get_subst \<tau> x = get_subst (subs i) x"
    proof (simp add: \<tau>, rule arg_cong[where f = "\<lambda> i. get_subst (subs i) x"])
      show "(THE i. i < ?n \<and> x \<in> vars_term (ls ! i)) = i"
      proof
        show "i < ?n \<and> x \<in> vars_term (ls ! i)" using i x by simp
      next
        fix j
        assume j: "j < ?n \<and> x \<in> vars_term (ls ! j)"
        show "j = i"
        proof (rule ccontr)
          assume neq: "\<not> j = i"
          from distinct[unfolded is_partition_alt is_partition_alt_def, THEN spec[of _ i], THEN spec[of _ j], THEN mp]
           i j x neq 
          show False by auto
        qed
      qed
    qed
  } note \<tau> = this
  show ?case
  proof (rule exI[of _ \<tau>], simp add: t, intro ballI conjI)
    fix li x
    assume li: "li \<in> set ls" and x: "x \<in> vars_term li"
    from li obtain i where i: "i < ?n" and li: "li = ls ! i" unfolding set_conv_nth by auto
    with x have x: "x \<in> vars_term (ls ! i)" by simp
    from subs[OF i, THEN conjunct2, THEN bspec, OF x] show "weak_match (get_subst \<tau> x) (get_subst \<sigma> x)" unfolding \<tau>[OF i x[unfolded li]]
      by simp
  next
    show "ts = map (\<lambda> t. t \<cdot> \<tau>) ls" 
    proof (rule nth_equalityI, simp add: lts, intro allI impI)
      fix i
      assume "i < length ts"
      with lts have i: "i < ?n" by simp
      have "ts ! i = ls ! i \<cdot> subs i"
        by (rule subs[THEN conjunct1, OF i])
      also have "\<dots> = ls ! i \<cdot> ((subs i)|_(vars_term(ls!i)))" (is "_ = _ \<cdot> ?subs")  by (rule coincidence_lemma)
      also have "\<dots> = ls ! i \<cdot> (\<tau>|_(vars_term(ls!i)))" (is "_ = _ \<cdot> ?tau") 
      proof (rule arg_cong[where f = "\<lambda> s. ls ! i \<cdot> s"])
        have subs: "Subst (get_subst (subs i)) = subs i" by (cases "subs i", simp)
        have tau: "Subst (get_subst \<tau>) = \<tau>" by (cases \<tau>, simp)        
        { 
          fix x
          have "get_subst ?subs x = get_subst ?tau x" (is "?l = ?r")
          proof (cases "x \<in> vars_term (ls ! i)")
            case False
            have "?l = get_subst ((Subst (get_subst (subs i))) |_(vars_term (ls ! i))) x" by (simp add: subs)
            also have "\<dots> = get_subst ((Subst (get_subst \<tau>)) |_(vars_term (ls ! i))) x" using False by simp
            also have "\<dots> = ?r" by (simp add: tau)
            finally show ?thesis .
          next
            case True
            have "?l = get_subst ((Subst (get_subst (subs i))) |_(vars_term (ls ! i))) x" by (simp add: subs)
            also have "\<dots> = get_subst (subs i) x" using True by simp
            also have "\<dots> = get_subst \<tau> x" using \<tau>[OF i True] by simp
            also have "\<dots> = get_subst ((Subst (get_subst \<tau>)) |_(vars_term (ls ! i))) x" using True by simp
            also have "\<dots> = ?r" by (simp add: tau)
            finally show ?thesis .
          qed
        } note main = this
        thus "?subs = ?tau" by (cases ?subs, cases ?tau, simp add: ext)
      qed
      also have "\<dots> = ls ! i \<cdot> \<tau>" by (rule coincidence_lemma[symmetric])
      finally show "ts ! i = map (\<lambda> t. t \<cdot> \<tau>) ls ! i"
        by (simp add: nth_map[OF i])
    qed
  qed
qed     


lemma subst_extend_flat_ctxt:
  assumes dist: "distinct vs"
    and len1: "length(take i (map Var vs)) = length ss1"
    and len2: "length(drop (Suc i) (map Var vs)) = length ss2"
    and i: "i < length vs"
  shows "More f (take i (map Var vs)) \<box> (drop (Suc i) (map Var vs)) \<cdot>c subst_extend \<sigma> (zip (take i vs@drop (Suc i) vs) (ss1@ss2)) = More f ss1 \<box> ss2"
proof -
  let ?V = "map Var vs"
  let ?vs1 = "take i vs" and ?vs2 = "drop (Suc i) vs"
  let ?ss1 = "take i ?V" and ?ss2 = "drop (Suc i) ?V"
  let ?\<sigma> = "subst_extend \<sigma> (zip (?vs1@?vs2) (ss1@ss2))"
  from len1 and len2 have len: "length(?vs1@?vs2) = length(ss1@ss2)" using i by simp
  from distinct_take_drop[OF dist i] have "distinct(?vs1@?vs2)" .
  from subst_extend_absorb[OF this len,of "\<sigma>"]
    have map: "map (\<lambda>t. t\<cdot>?\<sigma>) (?ss1@?ss2) = ss1@ss2" unfolding take_map drop_map map_append .
  from len1 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss1 = ss1" by auto
  moreover from len2 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss2 = ss2" by auto
  ultimately show ?thesis by simp
qed


lemma subst_extend_flat_ctxt'':
  assumes dist: "distinct vs"
    and len1: "length(take i (map Var vs)) = length ss1"
    and len2: "length(drop i (map Var vs)) = length ss2"
    and i: "i < length vs"
  shows "More f (take i (map Var vs)) \<box> (drop i (map Var vs)) \<cdot>c subst_extend \<sigma> (zip (take i vs@drop i vs) (ss1@ss2)) = More f ss1 \<box> ss2"
proof -
  let ?V = "map Var vs"
  let ?vs1 = "take i vs" and ?vs2 = "drop i vs"
  let ?ss1 = "take i ?V" and ?ss2 = "drop i ?V"
  let ?\<sigma> = "subst_extend \<sigma> (zip (?vs1@?vs2) (ss1@ss2))"
  from len1 and len2 have len: "length(?vs1@?vs2) = length(ss1@ss2)" using i by simp
  from distinct_take_drop'[OF dist i] have "distinct(?vs1@?vs2)" .
  from subst_extend_absorb[OF this len,of "\<sigma>"]
    have map: "map (\<lambda>t. t\<cdot>?\<sigma>) (?ss1@?ss2) = ss1@ss2" unfolding take_map drop_map map_append .
  from len1 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss1 = ss1" unfolding map_append by auto
  moreover from len2 and map have "map (\<lambda>t. t\<cdot>?\<sigma>) ?ss2 = ss2" unfolding map_append by auto
  ultimately show ?thesis by simp
qed

lemma distinct_map_Var: assumes "distinct xs" shows "distinct(map Var xs)"
using assms by (induct xs) auto

lemma terms_to_vars: assumes "list_all is_var ts" shows "\<Union>set(map vars_term ts) = set(map the_var ts)"
using assms proof (induct ts)
  case Nil thus ?case by simp
next
  case (Cons t ts)
  hence IH: "\<Union>set(map vars_term ts) = set(map the_var ts)" by simp
  from Cons have "is_var t" by simp
  hence "vars_term t = {the_var t}" by (induct t) auto
  with IH show ?case by simp
qed

lemma Var_the_var_id: assumes "list_all is_var ts" shows "map Var (map the_var ts) = ts"
using assms proof (induct ts)
  case Nil show ?case by simp
next
  case (Cons t ts)
  hence IH: "map Var (map the_var ts) = ts" and "is_var t" by auto
  from `is_var t` have "Var(the_var t) = t" by (induct t) auto
  with IH show ?case by simp
qed


lemma distinct_the_vars:
  assumes "list_all is_var ts" and "distinct ts" shows "distinct(map the_var ts)"
using assms proof (induct ts)
  case Nil show ?case by simp
next
  case (Cons t ts)
  hence IH: "distinct(map the_var ts)"
    and "distinct ts" and "list_all is_var ts" and "t \<notin> set ts" and "is_var t" by auto
  have "the_var t \<notin> set(map the_var ts)"
  proof
    assume "the_var t \<in> set(map the_var ts)"
    hence "Var(the_var t) \<in> set(map Var (map the_var ts))" unfolding map_map by auto
    moreover have "set(map Var (map the_var ts)) = set ts" using Var_the_var_id[OF `list_all is_var ts`] by simp
    ultimately have "Var(the_var t) \<in> set ts" by best
    with `is_var t` have "t \<in> set ts" by (induct t) auto
    with `t \<notin> set ts` show False by simp
  qed
  with IH show ?case by simp
qed

lemma funas_term_clean_subset: "funas_term (clean_term F t) \<subseteq> F" by (induct t) auto

lemma weak_match_map_term: "weak_match t s \<Longrightarrow> weak_match (map_term g t) (map_term g s)"
proof (induct s arbitrary: t, simp)
  case (Fun f ss t)
  from Fun(2) obtain ts where t: "t = Fun f ts" by (cases t, auto)
  from Fun(1)[unfolded set_conv_nth] Fun(2)[unfolded t]
  show ?case unfolding t 
    by force
qed

lemma vars_map_term[simp]: "vars_term o map_term f = vars_term"
proof
  fix t 
  show "(vars_term o map_term f) t = vars_term t"
    by (induct t, auto)
qed

lemma vars_map_term2[simp]: "vars_term (map_term f t) = vars_term t"
using fun_cong[OF vars_map_term, of f t] by (simp del: vars_map_term)

lemma linear_map_term[simp]: "linear_term (map_term f t) = linear_term t"
by (induct t, auto simp: list_all_iff)    


fun
  map_vars :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'w) term"
where
  "map_vars m (Var x) = (Var (m x))" |
  "map_vars m (Fun f ts) = Fun f (map (map_vars m) ts)"

lemma map_vars_compose:
  "map_vars m1 (map_vars m2 t) = map_vars (m1 o m2) t" 
by (induct t, auto)

lemma map_vars_id[simp]: "map_vars id t = t"
by (induct t) (auto intro: map_idI)

lemma map_vars_id'[simp]: "map_vars (\<lambda>x. x) t = t"
by (induct t) (auto intro: map_idI)

fun
  merge_substs :: "('f, 'v) subst \<Rightarrow> ('f, 'v) subst \<Rightarrow> ('f, 'v + 'v) subst"
where
  "merge_substs (Subst sigma) (Subst delta) = Subst (\<lambda>x.
     (case x of Inl y \<Rightarrow> map_vars Inl (sigma y)
              | Inr y \<Rightarrow> map_vars Inl (delta y)))"

lemma merge_substs_left:
  "map_vars Inl s \<cdot> (merge_substs \<sigma> \<delta>) = map_vars Inl (s \<cdot> \<sigma>)"
proof (induct s)
  case (Var x)
  thus ?case by (cases \<sigma>, cases \<delta>, auto)
next
  case (Fun f ts)
  thus ?case by auto
qed

lemma merge_substs_right:
  "map_vars Inr s \<cdot> (merge_substs \<sigma> \<delta>) = map_vars Inl (s \<cdot> \<delta>)"
proof (induct s)
  case (Var x)
  thus ?case by (cases \<sigma>, cases \<delta>, auto)
next
  case (Fun f ts)
  thus ?case by auto
qed

fun
  left_subst :: "('f, 'v + 'v) subst \<Rightarrow> ('f, 'v) subst"
where
  "left_subst (Subst sigma) = Subst (\<lambda>x. map_vars Sum_Type.Projl (sigma (Inl x)))"

fun
  right_subst :: "('f, 'v + 'v) subst \<Rightarrow> ('f, 'v) subst"
where
  "right_subst (Subst sigma) = Subst (\<lambda>x. map_vars Sum_Type.Projl (sigma (Inr x)))"

lemma left_subst:
  "map_vars Sum_Type.Projl (map_vars Inl s \<cdot> \<gamma>) =  s \<cdot> (left_subst \<gamma>)"
proof (induct s)
  case (Fun f ts) thus ?case by auto
next
  case (Var x) thus ?case by (cases \<gamma>, auto)
qed

lemma right_subst:
  "map_vars Sum_Type.Projl (map_vars Inr s \<cdot> \<gamma>) =  s \<cdot> (right_subst \<gamma>)"
proof (induct s)
  case (Fun f ts) thus ?case by auto
next
  case (Var x) thus ?case by (cases \<gamma>, auto)
qed

lemma size_subst: "size t \<le> size (t \<cdot> \<sigma>)"
proof (induct t)
  case (Var x)
  thus ?case by (cases \<sigma>, auto)
next
  case (Fun f ss)
  thus ?case 
    by (simp add: o_def, induct ss, force+)
qed

lemma Fun_Nil_supt[elim!]: "Fun f [] \<rhd> t \<Longrightarrow> P" by auto

lemma map_vars_vars_term: assumes "\<And> x. x \<in> vars_term t \<Longrightarrow> f x = g x"
  shows "map_vars f t = map_vars g t"
  using assms
proof (induct t, simp)
  case (Fun h ts)
  {
    fix t
    assume t: "t \<in> set ts"
    with Fun(2) have "\<And> x. x \<in> vars_term t \<Longrightarrow> f x = g x" 
      by auto 
    from Fun(1)[OF t this] have "map_vars f t = map_vars g t" by simp
  }
  thus ?case by auto
qed


lemma term_subst_eq: assumes "\<And> x. x \<in> vars_term t \<Longrightarrow> get_subst \<sigma> x = get_subst \<tau> x" 
  shows "t \<cdot> \<sigma> = t \<cdot> \<tau>"
  using assms by (induct t, auto)


lemma term_subst_eq_rev: "t \<cdot> \<sigma> = t \<cdot> \<tau> \<Longrightarrow> \<forall> x \<in> vars_term t. get_subst \<sigma> x = get_subst \<tau> x"
proof (induct t, simp)
  case (Fun f ts)
  thus ?case by auto
qed

lemma term_subst_eq_conv: "(t \<cdot> \<sigma> = t \<cdot> \<tau>) = (\<forall> x \<in> vars_term t. get_subst \<sigma> x = get_subst \<tau> x)"
  using term_subst_eq[of t \<sigma> \<tau>] term_subst_eq_rev[of t \<sigma> \<tau>] by blast

lemma map_term_ctxt_decomp: assumes "map_term fg t = C\<langle>s\<rangle>"
  shows "\<exists> D u. C = map_ctxt fg D \<and> s = map_term fg u \<and> t = D\<langle>u\<rangle>"
  using assms
proof (induct C arbitrary: t)
  case Hole
  show ?case
    by (rule exI[of _ Hole], rule exI[of _ t], insert Hole, auto)
next
  case (More g bef C aft)
  from More(2) obtain f ts where t: "t = Fun f ts" by (cases t, auto)
  from More(2)[unfolded t] have f: "fg f = g" and ts: "map (map_term fg) ts = bef @ C\<langle>s\<rangle> # aft" (is "?ts = ?bca") by auto
  from ts have "length ?ts = length ?bca" by auto
  hence len: "length ts = length ?bca" by auto
  note id = ts[unfolded map_nth_eq_conv[OF len], THEN spec, THEN mp]
  let ?i = "length bef"
  from len have i: "?i < length ts" by auto
  from id[of ?i] have "map_term fg (ts ! ?i) = C\<langle>s\<rangle>" by auto
  from More(1)[OF this] obtain D u where D: "C = map_ctxt fg D" and
    u: "s = map_term fg u" and id: "ts ! ?i = D\<langle>u\<rangle>" by auto
  from ts have "take ?i ?ts = take ?i ?bca" by simp
  also have "... = bef" by simp
  finally have bef: "map (map_term fg) (take ?i ts) = bef" by (simp add: take_map)
  from ts have "drop (Suc ?i) ?ts = drop (Suc ?i) ?bca" by simp
  also have "... = aft" by simp
  finally have aft: "map (map_term fg) (drop (Suc ?i) ts) = aft" by (simp add:drop_map)
  let ?bda = "take ?i ts @ D\<langle>u\<rangle> # drop (Suc ?i) ts"
  show ?case
  proof (rule exI[of _ "More f (take ?i ts) D (drop (Suc ?i) ts)"],
      rule exI[of _ u], simp add: u f D bef aft t)
    have "ts = take ?i ts @ ts ! ?i # drop (Suc ?i) ts"        
      by (rule id_take_nth_drop[OF i])
    also have "... = ?bda" by (simp add: id)
    finally show "ts = ?bda" .
  qed
qed


lemma ground_subst[simp]: "ground (t \<cdot> \<sigma>) = (\<forall> x \<in> vars_term t. ground (get_subst \<sigma> x))"
  by (induct t, auto)

fun ground_ctxt :: "('f,'v)ctxt \<Rightarrow> bool"
where "ground_ctxt Hole = True"
   |  "ground_ctxt (More f bef C aft) = (Ball (set bef) ground \<and> Ball (set aft) ground \<and> ground_ctxt C)"

lemma ground_ctxt_apply[simp]: "ground (C\<langle>t\<rangle>) = (ground_ctxt C \<and> ground t)"
  by (induct C, auto)

end
