
signature TRS_DATA =
sig
  val get_def : theory -> string -> (string * thm) option
  val get_trees : theory -> string -> (term * Call_Tree.tree) list
  val get_rules : theory -> string -> thm list
  val get_incl : theory -> string * string -> thm option
  val is_constr : theory -> string -> bool
  val add_constrs : string list -> theory -> theory

  val add_sim : thm -> Context.generic -> Context.generic
  val get_sims : Proof.context -> thm list
  val get_embs : Proof.context -> thm list
  val sim_for_trs : Proof.context -> string -> string -> thm

  val add_emb : attribute

  val define_trs : thm list -> theory -> theory
  val declare_constrs : string -> theory -> theory

  val setup : theory -> theory

  structure Embedding : NAMED_THMS
end

structure Trs_Data (*: TRS_DATA*) =
struct

fun duplicate_sim_err const_name =
  error ("Constant " ^ quote const_name ^ " does already have a simulation lemma");

open Isafor_Basics

structure Simulation = Generic_Data
(
  type T = thm Symtab.table
  val empty = Symtab.empty
  val extend = I
  val merge =
    Symtab.merge (K true)
    handle Symtab.DUP c => duplicate_sim_err c; (* FIXME? *)
)

fun const_for_sim thm =
  let val ((l,_),_) = dest_inrel (prop_of thm)
  in fst (dest_app l) end

fun add_sim thm = Simulation.map (Symtab.update_new (const_for_sim thm, thm))
val get_sims = map snd o Symtab.dest o Simulation.get o Context.Proof

val get_sim_for_const = Symtab.lookup o Simulation.get o Context.Proof
fun the_sim_for_const ctxt name =
  case get_sim_for_const ctxt name of
    SOME thm => thm
  | NONE => error ("No simulation lemma for " ^ quote name)

structure Embedding = 
  Named_Thms(val name = "emb"  val description = "emb rules");

structure Subst = 
  Named_Thms(val name = "trs_subst"  val description = "substitution rules for trs representation");

val get_embs = Embedding.get
val add_emb = Embedding.add

structure TrsData = Theory_Data(
  type T = (
    (string * thm) Symtab.table  (* deftab *)
  * ((term * Call_Tree.tree) list) Symtab.table (* treetab *)
  * (thm list) Symtab.table (* ruletab *)
  * thm Symreltab.table (* trs hierarchy inclusions *)
  * unit Symtab.table (* constructors *)
  )
  val empty = (Symtab.empty, Symtab.empty, Symtab.empty, Symreltab.empty, Symtab.empty)
  val extend = I
  fun merge ((x1,x2,x3,x4,x5), (y1,y2,y3,y4,y5)) = (
    Symtab.merge (K true) (x1,y1),
    Symtab.merge (K true) (x2,y2),
    Symtab.merge (K true) (x3,y3),
    Symreltab.merge (K true) (x4,y4),
    Symtab.merge (K true) (x5,y5))
)

val get_deftab = #1 o TrsData.get
val get_treetab = #2 o TrsData.get
val get_ruletab = #3 o TrsData.get
val get_incltab = #4 o TrsData.get
val get_constrs = #5 o TrsData.get

fun map_deftab f = TrsData.map (fn (a,b,c,d,e) => (f a, b, c, d, e))
fun map_treetab f = TrsData.map (fn (a,b,c,d,e) => (a, f b, c, d, e))
fun map_ruletab f = TrsData.map (fn (a,b,c,d,e) => (a, b, f c, d, e))
fun map_incltab f = TrsData.map (fn (a,b,c,d,e) => (a, b, c, f d, e))

fun add_constrs cs =
  TrsData.map (fn (a,b,c,d,e) => (a, b, c, d, fold (Symtab.update o rpair ()) cs e))

val get_def = Symtab.lookup o get_deftab
val get_trees = Symtab.lookup_list o get_treetab
fun get_rules thy name = Symtab.lookup_list (get_ruletab thy) name
val get_incl = Symreltab.lookup o get_incltab
val is_constr = Symtab.defined o get_constrs

fun sim_for_trs ctxt f_name trs_name = 
  let
    val sim = the_sim_for_const ctxt f_name
  in
    case dest_inrel (prop_of sim) of
      ((_, _), _ $ (_ $ Const (trs_name', _))) =>
        let val incl = the (get_incl (ProofContext.theory_of ctxt) (trs_name', trs_name))
        in @{thm subsetD} OF [incl, sim] end
    | _ => sim (* constructor sim rule *)
  end

fun set_member_tac m i =
  if m = 0 then rtac @{thm insertI1} i
  else rtac @{thm insertI2} i THEN set_member_tac (m - 1) i

fun extract_rules ctxt f defthm =
  let
    open Conv
    val thy = ProofContext.theory_of ctxt
    val (c, rhs) = Logic.dest_equals (prop_of defthm)
    val (rs :: _) = Function_Lib.dest_binop_list union_const_name rhs
    val rules = HOLogic.dest_set rs |> map_index I
      |> filter (fn (_, r) => 
        (r |> HOLogic.dest_prod |> fst |> dest_app |> fst) = f)
    fun prove_rule (m, r) =
      let 
        val cgoal = cterm_of thy (HOLogic.mk_Trueprop (HOLogic.mk_mem (r, c)))
      in
      Goal.prove_internal [] cgoal
        (K (CONVERSION (arg_conv (arg_conv (rewr_conv defthm))) 1
            THEN TRY (rtac @{thm UnI1} 1)
            THEN set_member_tac m 1))
      |> subst_closure ctxt
      |> (fn thm => thm RS @{thm CS_into_rstep})
      end
  in
    map prove_rule rules
  end

val Un_subsetI2 = @{lemma "A \<subseteq> C ==> A \<subseteq> B Un C" by auto}

fun extract_subsystems ctxt defthm =
  let
    open Conv
    val thy = ProofContext.theory_of ctxt
    val (c, rhs) = Logic.dest_equals (prop_of defthm)
    val (_ :: subs) = Function_Lib.dest_binop_list union_const_name rhs
    fun prove_sub (m, s) =
      let 
        val cgoal = HOLogic.mk_binrel @{const_name "Orderings.ord_class.less_eq"}
             (mk_star (mk_rstep s), mk_star (mk_rstep c))
          |> HOLogic.mk_Trueprop |> cterm_of thy
      in
        ((fst (dest_Const s), fst (dest_Const c)), Goal.prove_internal [] cgoal
          (K (rtac @{thm rtrancl_mono} 1
            THEN rtac @{thm rstep_mono} 1
            THEN CONVERSION (arg_conv (arg_conv (rewr_conv defthm))) 1
            THEN REPEAT_DETERM_N (m + 1) (rtac Un_subsetI2 1)
            THEN (rtac @{thm Un_upper1} 1 ORELSE rtac @{thm subset_refl} 1))))
      end
  in
    map_index prove_sub subs
  end

fun define_trs ceqns thy =
  let
    val eqns = map (Logic.strip_imp_concl o prop_of o Thm.unvarify_global) ceqns
    val fs = eqns
      |> map (fst o dest_Const o head_of o fst o HOLogic.dest_eq o HOLogic.dest_Trueprop)
      |> distinct (op =)
    val base_name = map Long_Name.base_name fs
      |> space_implode "_"
    val binding = Binding.name "trs"
      |> Binding.qualify true base_name

    val deftab = get_deftab thy
    fun trs_sym n = case Symtab.lookup deftab n of
      SOME (c, _) => Const (c, trsT)
    | NONE => error ("No TRS representation of " ^ quote n ^ ".")
    val treetab' = get_treetab thy
      |> fold Symtab.update_new (Call_Tree.analyze_rules fs (is_constr thy) thy eqns)
    val lhss_trees = map (Symtab.lookup_list treetab') fs
    val trees = maps (map snd) lhss_trees
    val case_syms = fold Call_Tree.add_case_syms trees [] |> rev
    val lhss_trees' = map (Symtab.lookup_list treetab') (fs @ case_syms)
    val trs = maps (map (apsnd Call_Tree.i_term)) lhss_trees'
      |> map HOLogic.mk_prod
      |> HOLogic.mk_set ruleT

    val sub_consts =
      fold Call_Tree.add_defined_syms (maps (map snd) lhss_trees') [] |> rev
      (*fold (Call_Tree.add_sim_syms base_name) (maps (map snd) lhss_trees') [] |> rev*)
      |> map trs_sym |> distinct (op aconv)
    val trs = foldr1 (HOLogic.mk_binop union_const_name)
      (trs :: sub_consts)

    val thy' = Sign.add_consts_i [(binding, trsT, NoSyn)] thy
    val Rf = Sign.full_name thy' binding
    val def = Logic.mk_equals (Const (Rf, trsT), trs)
    val ([defthm], thy'') = Global_Theory.add_defs false
       [((Binding.map_name (suffix "_def") binding, def), [])] thy'
    val ctxt'' = ProofContext.init_global thy''

    val inclusions = extract_subsystems ctxt'' defthm
    fun rules f = 
      (f, (extract_rules ctxt'' f defthm))
  in
    thy''
    |> map_deftab (fold (fn f => Symtab.update (f, (Rf, defthm))) fs)
    |> map_treetab (K treetab')
    |> map_ruletab (fold Symtab.update_new (map rules (fs @ case_syms)))
    |> map_incltab (fold Symreltab.update_new inclusions)
  end

fun declare_constrs tname thy =
let
  val constrs = the (Datatype.get_constrs thy tname)
in
  add_constrs (map fst constrs) thy
end


val setup = Embedding.setup
  #> Subst.setup
  #> Global_Theory.add_thms_dynamic (Binding.name "sim", map snd o Symtab.dest o Simulation.get)
  #> Global_Theory.add_thms_dynamic (Binding.name "all_trs_defs", fn context =>
       get_deftab (Context.theory_of context)
       |> Symtab.dest |> map snd |> AList.group (op =) |> map (hd o snd))

end
